# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
import datetime
import os
import pymast.predictors as predictors

def ares(file_name, 
                 db_dir, 
                 rec_id, 
                 study_tags, 
                 scan_time = 1, 
                 channels = 1, 
                 ant_to_rec_dict = None):
    # identify the receiver type 
    rec_type = 'ares'
    
    # what file format is it?  the header row is the key
    o_file =open(file_name, encoding='utf-8')
    header = o_file.readline()[:-1]                                            
    columns = header.split(',')
    o_file.close()
    
    # import ARES data and standardize
    if columns[0] == 'Date':
        telem_dat = pd.read_csv(file_name, 
                                names = ['Date','Time','RxID','Freq','Antenna',
                                         'Protocol','Code','Power','Squelch','Noise Level',
                                         'Pulse Width 1','Pulse Width 2','Pulse Width 3',
                                         'Pulse Interval 1','Pulse Interval 2','Pulse Interval 3'],
                                header = 0,
                                index_col = False)
        
        telem_dat['time_stamp'] = pd.to_datetime(telem_dat['Date'] + ' ' + telem_dat['Time'],errors = 'coerce')
        kHz = telem_dat.Freq * 1000
        kHz = np.round(kHz / 5.0, 0) * 5
        MHz = np.round(kHz / 1000.,3)
        telem_dat['Freq'] = MHz
        telem_dat['Freq'] = telem_dat['Freq'].apply(lambda x: f"{x:.3f}" )
        telem_dat['freq_code'] = telem_dat['Freq'].astype(str) + ' ' + telem_dat['Code'].astype(str)
        telem_dat.rename(columns = {'Power':'power'}, inplace = True)
        telem_dat.drop(columns = ['Date','Time','RxID','Freq','Antenna',
                                 'Protocol','Code','Squelch','Noise Level',
                                 'Pulse Width 1','Pulse Width 2','Pulse Width 3',
                                 'Pulse Interval 1','Pulse Interval 2','Pulse Interval 3'],
                       inplace = True)
    
    # import mitas data and standardize
    else:
        telem_dat = pd.read_csv(file_name,
                                names = ['contactId','decodeTimeUTC-05:00','ReceiverId',
                                         'antenna','frequency','codeNumber','protocol','power'],
                                header = 0,
                                index_col = None)
        
        telem_dat['decodeTimeUTC-05:00'] = pd.to_datetime(telem_dat['decodeTimeUTC-05:00'])
        kHz = telem_dat.frequency * 1000
        kHz = np.round(kHz / 5.0,0) * 5
        MHz = np.round(kHz / 1000., 3)
        telem_dat['frequency'] = MHz
        telem_dat['frequency'] = telem_dat['frequency'].apply(lambda x: f"{x:.3f}" )
        telem_dat['freq_code'] = telem_dat['frequency'].astype(str) + ' ' + telem_dat['codeNumber'].astype(str)
        telem_dat.rename(columns = {'decodeTimeUTC-05:00':'time_stamp'}, inplace = True)
        telem_dat.drop(columns = ['contactId','ReceiverId','antenna',
                                 'frequency','codeNumber','protocol'],
                       inplace = True)
        
    # now do this stuff to files regardless of type
    telem_dat['epoch'] = np.round((telem_dat.time_stamp - pd.Timestamp("1970-01-01")) / pd.Timedelta('1s'),6)
    telem_dat['rec_type'] = np.repeat(rec_type,len(telem_dat))
    telem_dat['rec_id'] = np.repeat(rec_id,len(telem_dat))
    telem_dat['channels'] = np.repeat(channels,len(telem_dat))
    telem_dat['scan_time'] = np.repeat(scan_time, len(telem_dat))
    telem_dat['noise_ratio'] = predictors.noise_ratio(5.0,
                                           telem_dat.freq_code.values,
                                           telem_dat.epoch.values,
                                           study_tags)
        
    telem_dat = telem_dat.astype({'power':'float32',
                                  'freq_code':'object',
                                  'time_stamp':'datetime64[ns]',
                                  'scan_time':'float32',
                                  'channels':'int32',
                                  'rec_type':'object',
                                  'epoch':'float32',
                                  'noise_ratio':'float32',
                                  'rec_id':'object'})
    
    with pd.HDFStore(db_dir, mode='a') as store:
        store.append(key = 'raw_data',
                     value = telem_dat, 
                     format = 'table', 
                     index = False,
                     min_itemsize = {'freq_code':20,
                                     'rec_type':20,
                                     'rec_id':20},
                     append = True, 
                     chunksize = 1000000)


def orion_import(file_name, 
                 db_dir, 
                 rec_id, 
                 study_tags, 
                 scan_time = 1, 
                 channels = 1, 
                 ant_to_rec_dict = None):
    '''Function imports raw Sigma Eight orion data.

    Text parser uses simple column fixed column widths.

    '''
    # identify the receiver type
    rec_type = 'orion'

    # what orion firmware is it?  the header row is the key
    o_file =open(file_name, encoding='utf-8')
    header = o_file.readline()[:-1]                                            
    columns = str.split(header)
    o_file.close()
    
    if 'Type' in columns:
        # with our data row, extract information using pandas fwf import procedure
        telem_dat = pd.read_fwf(file_name,colspecs = [(0,12),(13,23),(24,30),(31,35),(36,45),(46,54),(55,60),(61,65)],
                                names = ['Date','Time','Site','Ant','Freq','Type','Code','power'],
                                skiprows = 1,
                                dtype = {'Date':str,'Time':str,'Site':np.int32,'Ant':str,'Freq':str,'Type':str,'Code':str,'power':np.float64})
        telem_dat = telem_dat[telem_dat.Type != 'STATUS']
        telem_dat.drop(['Type'], axis = 1, inplace = True)

    else:
        # with our data row, extract information using pandas fwf import procedure
        telem_dat = pd.read_fwf(file_name,colspecs = [(0,11),(11,20),(20,26),(26,30),(30,37),(37,42),(42,48)],
                                names = ['Date','Time','Site','Ant','Freq','Code','power'],
                                skiprows = 1,
                                dtype = {'Date':str,'Time':str,'Site':str,'Ant':str,'Freq':str,'Code':str,'power':str})

    if len(telem_dat) > 0:
        # add file name to data
        #['fileName'] = np.repeat(file_name,len(telem_dat))    #Note I'm going back here to the actual file name without the path.  Is that OK?  I prefer it, but it's a potential source of confusion
        
        # aggregate frequency and code intoa unique identifier
        telem_dat['freq_code'] = telem_dat['Freq'].astype(str) + ' ' + telem_dat['Code'].astype(str)
        
        # concatenate date and time and create timestamp
        telem_dat['time_stamp'] = pd.to_datetime(telem_dat['Date'] + ' ' + telem_dat['Time'],errors = 'coerce')# create timestamp field from date and time and apply to index
        telem_dat['scan_time'] = np.repeat(scan_time,len(telem_dat))
        telem_dat['channels'] = np.repeat(channels,len(telem_dat))
        telem_dat['rec_type'] = np.repeat('orion',len(telem_dat))
        telem_dat = telem_dat[telem_dat.time_stamp.notnull()]
        
        if len(telem_dat) == 0:
            print ("Invalid timestamps in raw data, cannot import")
        else:
            # create epoch
            telem_dat['epoch'] = np.round((telem_dat.time_stamp - pd.Timestamp("1970-01-01")) / pd.Timedelta('1s'),6)
            
            # drop unnecessary columns 
            telem_dat.drop (['Date','Time','Freq','Code','Site'],axis = 1, inplace = True)
            
            # calculate noise ratio
            telem_dat['noise_ratio'] = predictors.noise_ratio(5.0,
                                                   telem_dat.freq_code.values,
                                                   telem_dat.epoch.values,
                                                   study_tags)
            
            # if there is no antenna to receiver dictionary 
            if ant_to_rec_dict == None:
                # drop the antenna column - we don't need it anymore
                telem_dat.drop(['Ant'], axis = 1, inplace = True)
                
                # add receiver id 
                telem_dat['rec_id'] = np.repeat(rec_id,len(telem_dat))

                telem_dat = telem_dat.astype({'power':'float32',
                                              'freq_code':'object',
                                              'time_stamp':'datetime64[ns]',
                                              'scan_time':'float32',
                                              'channels':'int32',
                                              'rec_type':'object',
                                              'epoch':'float32',
                                              'noise_ratio':'float32',
                                              'rec_id':'object'})
                
                telem_dat = telem_dat[['power', 
                                        'time_stamp',
                                        'epoch',
                                        'freq_code',
                                        'noise_ratio',
                                        'scan_time',
                                        'channels', 
                                        'rec_id',
                                        'rec_type']]
                
                with pd.HDFStore(db_dir, mode='a') as store:
                    store.append(key = 'raw_data',
                                 value = telem_dat, 
                                 format = 'table', 
                                 index = False, 
                                 min_itemsize = {'freq_code':20,
                                                 'rec_type':20,
                                                 'rec_id':20},
                                 append = True, 
                                 chunksize = 1000000,
                                 data_columns = True)  
                
            # if there is an antenna to receiver dictionary
            else:
                for i in ant_to_rec_dict:
                    # get site from dictionary
                    site = ant_to_rec_dict[i]
                    
                    # get telemetryt data associated with this site
                    telem_dat_sub = telem_dat[telem_dat.Ant == str(i)]
                    
                    # add receiver ID
                    telem_dat_sub['rec_id'] = np.repeat(site,len(telem_dat_sub))
                    
                    # remove exctranneous columns
                    telem_dat_sub.drop(['Ant'], axis = 1, inplace = True)
                    
                    telem_dat_sub = telem_dat_sub.astype({'power':'float32',
                                                          'freq_code':'object',
                                                          'time_stamp':'datetime64[ns]',
                                                          'scan_time':'float32',
                                                          'channels':'int32',
                                                          'rec_type':'object',
                                                          'epoch':'float32',
                                                          'noise_ratio':'float32',
                                                          'rec_id':'object'})
                    
                    telem_dat_sub = telem_dat_sub[['power',
                                                   'time_stamp',
                                                   'epoch', 
                                                   'freq_code', 
                                                   'noise_ratio',
                                                   'scan_time', 
                                                   'channels',
                                                   'rec_id',
                                                   'rec_type']]
                    
                    with pd.HDFStore(db_dir, mode='a') as store:
                        store.append(key = 'raw_data',
                                     value = telem_dat_sub, 
                                     format = 'table', 
                                     index = False, 
                                     min_itemsize = {'freq_code':20,
                                                     'rec_type':20,
                                                     'rec_id':20},
                                     append = True, 
                                     chunksize = 1000000,
                                     data_columns = True)  

                    
                    
def vr2_import(file_name,db_dir,study_tags, rec_id):
    '''Function imports raw VEMCO VR2 acoustic data.

    '''

    recType = 'vr2'

    # import csv format and export - pretty simple
    telem_dat = pd.read_csv(file_name)

    # if data isn't empty - hey, things go wrong
    if len(telem_dat) > 0:
        # depending upon header data create a timestep
        try:
            telem_dat['time_stamp'] = pd.to_datetime(telem_dat['Date and Time (UTC)'])
        except KeyError:
            seconds = pd.to_numeric(telem_dat['Unnamed: 2'].str.split(":", n = 1, expand = True)[1])
            telem_dat['time_stamp'] = pd.to_datetime(telem_dat['Date and Time'])
            telem_dat['time_stamp'] = telem_dat.timeStamp + pd.to_timedelta(seconds,unit = 's')
            telem_dat['rec_id'] = telem_dat['Receiver'].str.split("-", expand = True)[0]
            telem_dat['rec_id'] = telem_dat.Receiver + "-" + telem_dat['Receiver.1'].astype('str')
        telem_dat['scan_time'] = np.repeat(1,len(telem_dat))
        telem_dat['channels'] = np.repeat(1,len(telem_dat))
        telem_dat['rec_type'] = np.repeat('vr2',len(telem_dat))
        telem_dat['transmitter'] = telem_dat['transmitter'].str.split("-", n = 2, expand = True)[2]
        telem_dat['transmitter'] = telem_dat.transmitter.astype(str)
        telem_dat.rename(columns = {'Receiver':'rec_id','transmitter':'freq_code'}, inplace = True)
        telem_dat['epoch'] = np.round((telem_dat.time_stamp - pd.Timestamp("1970-01-01")) / pd.Timedelta('1s'),6)
        try:
            telem_dat.drop (['Date and Time (UTC)', 'Transmitter Name','Transmitter Serial','Sensor Value','Sensor Unit','Station Name','Latitude','Longitude','Transmitter Type','Sensor Precision'],axis = 1, inplace = True)
        except KeyError:
            telem_dat.drop (['Unnamed: 0', 'Date and Time', 'Unnamed: 2', 'Unnamed: 3', 'Unnamed: 4', 'Receiver.1', 'Unnamed: 7', 'Unnamed: 8', 'Transmitter Name', 'Unnamed: 11', 'Unnamed: 12', 'Unnamed: 13',],axis = 1, inplace = True)

        # tuples = zip(telem_dat.FreqCode.values,telem_dat.recID.values,telem_dat.Epoch.values)
        # index = pd.MultiIndex.from_tuples(tuples, names=['freq_code', 'rec_id','Epoch'])
        # telem_dat.set_index(index,inplace = True,drop = False)
        
        telem_dat = telem_dat.astype({'power':'float32',
                                      'freq_code':'object',
                                      'time_stamp':'datetime64[ns]',
                                      'scan_time':'float32',
                                      'channels':'int32',
                                      'rec_type':'object',
                                      'epoch':'float32',
                                      'noise_ratio':'float32',
                                      'rec_id':'object'})
        
        with pd.HDFStore(db_dir, mode='a') as store:
            store.append(key = 'raw_data',
                         value = telem_dat, 
                         format = 'table', 
                         index = False,
                         min_itemsize = {'freq_code':20,
                                         'rec_type':20,
                                         'rec_id':20},
                         append = True, 
                         chunksize = 1000000)        
def srx1200(file_name,
             db_dir,
             rec_id, 
             study_tags, 
             scan_time = 1, 
             channels = 1, 
             ant_to_rec_dict = None):
    rec_type = 'srx1200'
    
    # create empty dictionary to hold Lotek header data indexed by line number - to be imported to Pandas dataframe
    header_dat = {}
    
    # create empty array to hold line indices
    line_counter = []
    
    # generate a list of header lines - contains all data we need to write to project set up database
    line_list = []
    
    # open file passed to function
    o_file = open(file_name, encoding='utf-8')
    
    # start line counter
    counter = 0
    
    # start end of file 
    eof = 0
    line_counter.append(counter)
    
    # read first line in file
    line = o_file.readline()[:-1]
    
    # append the current line of header data to the line list
    line_list.append(line)
    
    # determine file format 
    file_header = True
    if 'SRX1200 Data File Export' not in line:
        file_header = False
    
    # if the srx1200 file has header information
    if file_header == True:
        # find where data begins and header data ends
        with o_file as f:
            for line in f:
                # if this current line signifies the start of the data stream, the data starts three rows down from this
                if "All Detection Records" in line:
                    counter = counter + 1
                    dataRow = counter + 5
                    # break the loop, we have reached our stop point
                    
                # else we are still reading header data increase the line counter by 1
                elif "Battery Report" in line:
                    counter = counter + 1
                    dataEnd = counter - 2
                else:
                    counter = counter + 1
                    # append the line counter to the count array
                    line_counter.append(counter)
                    # append line of data to the data array
                    line_list.append(line)
                    eof = counter

        # add count array to dictionary with field name 'idx' as key
        header_dat['idx'] = line_counter
        # add data line array to dictionary with field name 'line' as key
        header_dat['line'] = line_list
        # create pandas dataframe of header data indexed by row number
        header_df = pd.DataFrame.from_dict(header_dat)
        header_df.set_index('idx',inplace = True)
            
        # get setups, scan times, firmware, and active scan tables
        setups = {}
        scan_times = {}
        firmware = ''
        setup_no = 0
        scan_channel = []
        scan_table_setups = {}
        scan_table_number = 0
        scan_time_number = 0
        new_split = None
        for row in header_df.iterrows():
            if 'SCAN SETTINGS' in row[1][0]:
                next_row = header_df.loc[row[0]+1]
                scanTimeStr = next_row.loc['line'][-7:-1]
                scanTimeSplit = scanTimeStr.split(':')
                if 'sec' in scanTimeSplit[1]:
                    new_split = scanTimeSplit[1].split(' ')
                    scanTimeSplit[1] = new_split[8]
                else:
                    scan_time = float(scanTimeSplit[1])
                    scan_times[scan_time_number] = scan_time
                    scan_time_number = scan_time_number + 1
                    
            if 'Environment History:' in row[1][0]:
                next_row = header_df.loc[row[0]+1]
                change_str = next_row.loc['line']
                setups[setup_no] = change_str[8:-1]
                setup_no = setup_no + 1
            
            if 'Master Firmware:' in row[1][0]:
                firmwarestr = row[1][0]
                firmware_split = firmwarestr.split(' ')
                firmware = str(firmware_split[-1]).lstrip().rstrip()
                
            if 'Active scan_table:' in row[1][0]:
                idx0 = row[0] + 2
                row_count = 1
                next_row = header_df.loc[row[0] + 1]
                while next_row[0] != '\n':
                    row_count = row_count + 1
                    next_row = header_df.loc[row[0] + row_count]
                    
                idx1 = idx0 + row_count - 2
                scan_table_setups[scan_table_number] = (idx0,idx1)
                scan_table_number = scan_table_number + 1
                
        scan_times_df = pd.DataFrame.from_dict(scan_times,
                                               orient = 'index',
                                               columns = ['scan_time'])
        scan_times_df.reset_index(inplace = True, drop = False)
        scan_times_arr = np.column_stack((scan_times_df.index, scan_times_df.scan_time))
        
        setup_df = pd.DataFrame.from_dict(setups, 
                                          orient = 'index', 
                                          columns = ['change_date'])
        setup_df['change_date'] = pd.to_datetime(setup_df.change_date)
        setup_df['epoch'] = (setup_df.change_date - pd.Timestamp("1970-01-01")) / pd.Timedelta('1s')        
        setup_df.reset_index(inplace = True, drop = False)
        setup_df.rename(columns = {'index':'setup'}, inplace = True)
    
        # extract channel dictionary data using rows identified 
        channel_dict = {}
        idx = 0
        for setup in scan_table_setups:
            idx0 = scan_table_setups[setup][0]
            idx1 = scan_table_setups[setup][1]
            channel_dat = header_df.iloc[idx0:idx1]
            for row in channel_dat.iterrows():
                dat = row[1][0]
                channel = str(dat[0:4]).lstrip()
                frequency = dat[10:17]
                channel_dict[idx] = [setup, channel, frequency]
                idx = idx + 1
            
        freq_scan_df = pd.DataFrame.from_dict(channel_dict, 
                                              orient = 'index', 
                                              columns = ['setup','channel','frequency'])
        # convert to numpy for vectorizing 
        freq_scan_arr = np.column_stack((freq_scan_df.setup.values,
                                         freq_scan_df.channel.values,
                                         freq_scan_df.frequency.values))
        
        print ('parsing SRX1200 header complete')

        # read in telemetry data
        if new_split == None:
            telem_dat = pd.read_fwf(file_name,
                                   colspecs = [(0,6),(6,20),(20,32),(32,40),(40,55),(55,64),(64,72),(72,85),(85,93),(93,102)],
                                   names = ['Index','Date','Time','[uSec]','Tag/BPM','Freq [MHz]','Codeset','Antenna','Gain','RSSI'],
                                   skiprows = dataRow, 
                                   skipfooter = eof - dataEnd)
            telem_dat.drop(columns = ['Index'], inplace = True)

        else:
            telem_dat = pd.read_csv(file_name,
                                   names = ['Index','Date','Time','[uSec]','Tag/BPM','Freq [MHz]','Codeset','Antenna','','Gain','RSSI'],
                                   skiprows = dataRow, 
                                   skipfooter = eof - dataEnd)
            telem_dat.drop(columns = ['', 'Index'], inplace = True)
            telem_dat.dropna(inplace = True)
            telem_dat = telem_dat.astype({'[uSec]':'int32'})

        
        # create a timestamp and conver to datetime object
        telem_dat['time_stamp'] = telem_dat.Date + ' ' + telem_dat.Time + '.' + telem_dat['[uSec]'].astype(str)

        telem_dat['time_stamp'] = pd.to_datetime(telem_dat.time_stamp)
        
        # calculate Epoch
        telem_dat['epoch'] = (telem_dat.time_stamp - pd.Timestamp("1970-01-01")) / pd.Timedelta('1s')
        
        # format frequency code
        telem_dat['FreqNo'] = telem_dat['Freq [MHz]'].apply(lambda x: f"{x:.3f}" )
        telem_dat = telem_dat[telem_dat['Tag/BPM'] != 999]
        telem_dat['freq_code'] = telem_dat['FreqNo'] + ' ' + telem_dat['Tag/BPM'].astype(str)
        
        # calculate 
        telem_dat['noise_ratio'] = predictors.noise_ratio(600,
                                              telem_dat.freq_code.values,
                                              telem_dat.epoch.values,
                                              study_tags)

        # write scan time, channels, rec type and recID to data
        telem_dat['scan_time'] = np.repeat(scan_time,len(telem_dat))
        telem_dat['channels'] = np.repeat(channels,len(telem_dat))
        telem_dat['rec_type'] = np.repeat(rec_type,len(telem_dat))
        telem_dat['rec_id'] = np.repeat(rec_id,len(telem_dat))
        telem_dat['noise_ratio'] = telem_dat.noise_ratio.values.astype(np.float32)
       
        # remove unnecessary columns
        telem_dat.drop(columns = ['Date','Time','[uSec]','Freq [MHz]','FreqNo', 'Codeset', 'Gain','Tag/BPM','Antenna'], inplace = True)
        telem_dat.rename(columns = {'RSSI':'power'}, inplace = True)
        telem_dat['power'] = telem_dat.power.values.astype(np.float32)
        telem_dat['noise_ratio'] = telem_dat.noise_ratio.values.astype(np.float32)

        telem_dat.reset_index(inplace = True)
        
        telem_dat = telem_dat.astype({'power':'float32',
                                      'freq_code':'object',
                                      'time_stamp':'datetime64[ns]',
                                      'scan_time':'int32',
                                      'channels':'int32',
                                      'rec_type':'object',
                                      'epoch':'float32',
                                      'noise_ratio':'float32',
                                      'rec_id':'object'})
        
        if new_split != None:
            telem_dat.drop(columns = ['index'], inplace = True)
            print ('fuck')
                
        with pd.HDFStore(db_dir, mode='a') as store:
            store.append(key = 'raw_data',
                         value = telem_dat, 
                         format = 'table', 
                         index = False, 
                         append = True, 
                         min_itemsize = {'freq_code':20,
                                         'rec_type':20,
                                         'rec_id':20},
                         chunksize = 1000000, 
                         data_columns = True,)
        
    # if the data doesn't have a header
    else:
        stop = 0
        eof = 0
        # find where data ends
        with o_file as f:
            for line in f:
                if "Battery Report" in line:
                    stop = counter - 1

                else:
                    counter = counter + 1
                    eof = counter
        
        # import data using specs defined above
        telem_dat = pd.read_csv(file_name, skiprows = 3, skipfooter = eof - stop)
        
        # create a timestamp and conver to datetime object
        telem_dat['time_stamp'] = telem_dat.Date + ' ' + telem_dat.Time + '.' + telem_dat['[uSec]'].astype(str)
        telem_dat['time_stamp'] = pd.to_datetime(telem_dat.time_stamp)
        
        # calculate Epoch
        telem_dat['epoch'] = np.round((telem_dat.time_stamp - pd.Timestamp("1970-01-01")) / pd.Timedelta('1s'),6)
                
        # format frequency code
        telem_dat['FreqNo'] = telem_dat['Freq [MHz]'].apply(lambda x: f"{x:.3f}" )
        telem_dat = telem_dat[telem_dat['TagID/BPM'] != 999]

        telem_dat['freq_code'] = telem_dat['FreqNo'] + ' ' + telem_dat['TagID/BPM'].astype(str)
        
        # calculate 
        telem_dat['noise_ratio'] = predictors.noise_ratio(600,
                                              telem_dat.freq_code.values,
                                              telem_dat.epoch.values,
                                              study_tags)

        # write scan time, channels, rec type and recID to data
        telem_dat['scan_time'] = np.repeat(scan_time,len(telem_dat))
        telem_dat['channels'] = np.repeat(channels,len(telem_dat))
        telem_dat['rec_type'] = np.repeat(rec_type,len(telem_dat))
        telem_dat['rec_id'] = np.repeat(rec_id,len(telem_dat))
                
        # remove unnecessary columns
        telem_dat.drop(columns = ['Index','Date','Time','[uSec]','Freq [MHz]','FreqNo', 'Codeset', 'Gain','TagID/BPM','Antenna'], inplace = True)
        telem_dat.rename(columns = {'RSSI':'power'}, inplace = True)
        telem_dat['noise_ratio'] = telem_dat.noise_ratio.values.astype(np.float32)
        telem_dat['power'] = telem_dat.power.values.astype(np.float32)
        
        telem_dat.reset_index(inplace = True, drop = True)
        
        telem_dat = telem_dat.astype({'power':'float32',
                                      'freq_code':'object',
                                      'time_stamp':'datetime64[ns]',
                                      'scan_time':'float32',
                                      'channels':'int32',
                                      'rec_type':'object',
                                      'epoch':'float32',
                                      'noise_ratio':'float32',
                                      'rec_id':'object'})
                
        with pd.HDFStore(db_dir, mode='a') as store:
            store.append(key = 'raw_data',
                         value = telem_dat, 
                         format = 'table', 
                         index = False, 
                         min_itemsize = {'freq_code':20,
                                         'rec_type':20,
                                         'rec_id':20},
                         append = True, 
                         chunksize = 1000000,
                         data_columns = True)   
            
def srx800(file_name,
             db_dir,
             rec_id, 
             study_tags, 
             scan_time = 1, 
             channels = 1, 
             ant_to_rec_dict = None):
    
    rec_type = 'srx800'
    
    # create empty dictionary to hold Lotek header data indexed by line number - to be imported to Pandas dataframe
    header_dat = {}
    
    # create empty array to hold line indices
    line_counter = []
    
    # generate a list of header lines - contains all data we need to write to project set up database
    line_list = []
    
    # open file passed to function
    o_file = open(file_name, encoding='utf-8')
    
    # start line counter
    counter = 0
    
    # start end of file 
    eof = 0
    line_counter.append(counter)
    
    # read first line in file
    line = o_file.readline()[:-1]
    
    # append the current line of header data to the line list
    line_list.append(line)
    
    # find where data begins and header data ends
    with o_file as f:
        for line in f:
            # if this current line signifies the start of the data stream, the data starts three rows down from this
            if "** Data Segment **" in line:
                counter = counter + 1
                dataRow = counter + 5

            # else we are still reading header data increase the line counter by 1
            else:
                counter = counter + 1
                # append the line counter to the count array
                line_counter.append(counter)
                # append line of data to the data array
                line_list.append(line)
                eof = counter

    # add count array to dictionary with field name 'idx' as key
    header_dat['idx'] = line_counter
    # add data line array to dictionary with field name 'line' as key
    header_dat['line'] = line_list
    # create pandas dataframe of header data indexed by row number
    headerDF = pd.DataFrame.from_dict(header_dat)
    headerDF.set_index('idx',inplace = True)
    
    data_format = 'primary'
    if 'Environment History:' in headerDF.at[0,'line']:
        data_format = 'alternate'

    # get setups, scan times, firmware, and active scan tables
    setups = {}
    scan_times = {}
    firmware = ''
    setup_no = 0
    scan_channel = []
    scan_table_setups = {}
    scan_table_number = 0
    scan_time_number = 0
    for row in headerDF.iterrows():
        if 'Environment History:' in row[1][0]:
            next_row = headerDF.loc[row[0]+1]
            change_str = next_row.loc['line']
            setups[setup_no] = change_str[8:-1]
            setup_no = setup_no + 1
        
        if 'Master Firmware:' in row[1][0]:
            firmwarestr = row[1][0]
            firmware_split = firmwarestr.split(' ')
            firmware = str(firmware_split[-1]).lstrip().rstrip()
            
        if 'SCAN SETTINGS:' in row[1][0]:
            next_row = headerDF.loc[row[0]+1]
            scanTimeStr = next_row.loc['line'][-7:-1]
            scanTimeSplit = scanTimeStr.split(':')
            scan_time = float(scanTimeSplit[1])
            scan_times[scan_time_number] = scan_time
            scan_time_number = scan_time_number + 1  
            
        if 'scan time' in row[1][0]:
            scanTimeStr = row[1]['line'][-7:-1]
            scanTimeSplit = scanTimeStr.split(':')
            scan_time = float(scanTimeSplit[1])
            scan_times[scan_time_number] = scan_time
            scan_time_number = scan_time_number + 1             
            
        if 'Active scan_table:' in row[1][0]:
            idx0 = row[0] + 2
            row_count = 1
            next_row = headerDF.loc[row[0] + 1]
            while next_row[0] != '\n':
                row_count = row_count + 1
                next_row = headerDF.loc[row[0] + row_count]
                
            idx1 = idx0 + row_count - 2
            scan_table_setups[scan_table_number] = (idx0,idx1)
            scan_table_number = scan_table_number + 1
            
    scan_times_df = pd.DataFrame.from_dict(scan_times,
                                           orient = 'index',
                                           columns = ['scan_time'])
    scan_times_df.reset_index(inplace = True, drop = False)
    scan_times_arr = np.column_stack((scan_times_df.index, scan_times_df.scan_time))
    
    setup_df = pd.DataFrame.from_dict(setups, 
                                      orient = 'index', 
                                      columns = ['change_date'])
    if data_format == 'alternate':
        split = setup_df['change_date'].str.split(' ', expand=True)
        setup_df['day0'] = np.repeat(pd.to_datetime("1900-01-01"),len(setup_df))
        setup_df['Date'] = setup_df['day0'] + pd.to_timedelta(split[1].astype(int), unit='d')
        setup_df['change_date'] = setup_df.Date.astype(str) + ' ' + split[2]
        

    setup_df['change_date'] = pd.to_datetime(setup_df.change_date)
    setup_df['epoch'] = (setup_df['change_date'] - datetime.datetime(1970,1,1)).dt.total_seconds()
    setup_df.reset_index(inplace = True, drop = False)
    setup_df.rename(columns = {'index':'setup'}, inplace = True)

    # extract channel dictionary data using rows identified 
    channel_dict = {}
    idx = 0
    for setup in scan_table_setups:
        idx0 = scan_table_setups[setup][0]
        idx1 = scan_table_setups[setup][1]
        channel_dat = headerDF.iloc[idx0:idx1]
        for row in channel_dat.iterrows():
            dat = row[1][0]
            channel = str(dat[0:4]).lstrip()
            frequency = dat[10:17]
            channel_dict[idx] = [setup, channel, frequency]
            idx = idx + 1
        
    freq_scan_df = pd.DataFrame.from_dict(channel_dict, 
                                          orient = 'index', 
                                          columns = ['setup','channel','frequency'])
    # convert to numpy for vectorizing 
    freq_scan_arr = np.column_stack((freq_scan_df.setup.values,
                                     freq_scan_df.channel.values,
                                     freq_scan_df.frequency.values))
    
    print ('parsing SRX800 header complete')
    
    # with our data row, extract information using pandas fwf import procedure
    if firmware == '9.12.5':
        telem_dat = pd.read_fwf(file_name,
                               colspecs = [(0,8),(8,23),(23,33),(33,41),(41,56),(56,64)],
                               names = ['Date','Time','ChannelID','TagID','Antenna','power'],
                               skiprows = dataRow,
                               dtype = {'ChannelID':str,'TagID':str,'Antenna':str})
    elif data_format == 'alternate':
        telem_dat = pd.read_fwf(file_name,
                               colspecs = [(0,5),(5,14),(14,23),(23,31),(31,46),(46,53)],
                               names = ['DayNumber','Time','ChannelID','TagID','Antenna','power'],
                               skiprows = dataRow,
                               dtype = {'ChannelID':str,'TagID':str,'Antenna':str})
        telem_dat['day0'] = np.repeat(pd.to_datetime("1900-01-01"),len(telem_dat))
        telem_dat['Date'] = telem_dat['day0'] + pd.to_timedelta(telem_dat['DayNumber'].astype(int), unit='d')
        telem_dat['Date'] = telem_dat.Date.astype('str')
        telem_dat.drop(columns = ['DayNumber','day0'], inplace = True)
    else:
        telem_dat = pd.read_fwf(file_name,
                               colspecs = [(0,8),(8,18),(18,28),(28,36),(36,51),(51,59)],
                               names = ['Date','Time','ChannelID','TagID','Antenna','power'],
                               skiprows = dataRow,
                               dtype = {'ChannelID':str,'TagID':str,'Antenna':str})

    telem_dat.dropna(inplace = True)
    print ('data import complete')
    
    # create some vectorizable functions that get setups, frequencies, channels, and scan times
    def setup_class(curr_date, setups):
        # Find indices where the condition is True
        indices = np.where(curr_date > setups)
        
        # Get the last index that meets the condition
        last_index = indices[0][-1] if indices[0].size > 0 else setups.shape[0]-1

        return last_index
    
    get_setup = np.vectorize(setup_class, excluded = [1])
    
    def freq_class(curr_setup, curr_channel, freq_scan_arr):
        # Condition: input_value1 > column 1 AND input_value2 < column 2
        condition = (freq_scan_arr[:, 0] == curr_setup) & (freq_scan_arr[:, 1] == curr_channel)
        
        # Find indices where the condition is True
        indices = np.where(condition)
        
        freq = freq_scan_arr[indices][:, 2][0]
        return freq
            
    get_frequency = np.vectorize(freq_class, excluded = [2])
    
    def count_channels(curr_setup, freq_scan_arr):
        condition = freq_scan_arr[:, 0] == curr_setup
        channels = condition.sum()
        return channels
    
    get_channels = np.vectorize(count_channels, excluded = [1])
    
    def scan_time_class(curr_setup, scan_times_arr):
        condition = scan_times_arr[:, 0] == curr_setup
        indices = np.where(condition)
        scan_time = scan_times_arr[indices][:, 1][0]
        return scan_time
        
    get_scan_time = np.vectorize(scan_time_class, excluded = [1])
    
    if len(telem_dat) > 0:
        for ant in ant_to_rec_dict:
            site = ant_to_rec_dict[ant]
            telem_dat_sub = telem_dat[telem_dat.Antenna == ant]
            
            # create timestamp
            telem_dat_sub['time_stamp'] = pd.to_datetime(telem_dat_sub['Date'] + ' ' + telem_dat_sub['Time'])
            
            # get UNIX epoch
            telem_dat_sub['epoch'] = np.round((telem_dat_sub.time_stamp - pd.Timestamp("1970-01-01")) / pd.Timedelta('1s'),6)
            
            # get setup number for every row
            try:
                telem_dat_sub['setup'] = get_setup(telem_dat_sub.epoch.values,
                                                   setup_df.epoch.values)
            except:
                print ('why you fail?')
            
            # get frequency from channel
            telem_dat_sub['Frequency'] = get_frequency(telem_dat_sub.setup.values,
                                                       telem_dat_sub.ChannelID.values,
                                                       freq_scan_arr)

            # remove extraneous data
            telem_dat_sub = telem_dat_sub[telem_dat_sub.Frequency != '888']
            #telemDat_sub = telemDat_sub[telemDat_sub.TagID != 999]
            
            # create FreqCode
            telem_dat_sub['freq_code'] = telem_dat_sub['Frequency'].astype(str) + ' ' + telem_dat_sub['TagID'].astype(int).astype(str)
            
            if np.any(pd.isnull(telem_dat_sub.Frequency.values)):
                print ('debug - why is frequency not working?')
            
            # calculate noise ratio
            telem_dat_sub['noise_ratio'] = predictors.noise_ratio(5.0,
                                                                  telem_dat_sub.freq_code,
                                                                  telem_dat_sub.epoch,
                                                                  study_tags)
            
            # write scan time, channels, rec type and recID to data
            telem_dat_sub['scan_time'] = get_scan_time(telem_dat_sub.setup.values, scan_times_arr)
            telem_dat_sub['channels'] = get_channels(telem_dat_sub.setup.values,freq_scan_arr)
            telem_dat_sub['scan_time'] = np.where(telem_dat_sub.channels == 1, 1, scan_time)
            telem_dat_sub['rec_id'] = np.repeat(site,len(telem_dat_sub))
            telem_dat_sub['rec_type'] = np.repeat(rec_type,len(telem_dat_sub))

            
            # remove extraneous columns
            telem_dat_sub.drop (['Date','Time','Frequency','TagID','ChannelID','Antenna','setup'],axis = 1, inplace = True)
            telem_dat_sub.reset_index(inplace = True)
            
            # write to hdf
            telem_dat_sub = telem_dat_sub.astype({'power':'float32',
                                                  'freq_code':'object',
                                                  'time_stamp':'datetime64[ns]',
                                                  'scan_time':'float32',
                                                  'channels':'int32',
                                                  'rec_type':'object',
                                                  'epoch':'float32',
                                                  'noise_ratio':'float32',
                                                  'rec_id':'object'})
                
            telem_dat_sub.reset_index(inplace = True, drop = True)
            
            telem_dat_sub = telem_dat_sub[['power', 
                                           'time_stamp',
                                           'epoch',
                                           'freq_code',
                                           'noise_ratio',
                                           'scan_time',
                                           'channels', 
                                           'rec_id',
                                           'rec_type']]
            
            telem_dat_sub.to_hdf(db_dir,
                                 key = 'raw_data',
                                 mode = 'a',
                                 append = True,
                                 format = 'table',
                                 index = False,
                                 min_itemsize = {'freq_code':20,
                                                 'rec_type':20,
                                                 'rec_id':20},
                                 data_columns = True)
            

            print ('Data standardized and exported to hdf')

def srx600(file_name,
             db_dir,
             rec_id, 
             study_tags, 
             scan_time = 1, 
             channels = 1, 
             ant_to_rec_dict = None):
    
    rec_type = 'srx600'
    
    # create empty dictionary to hold Lotek header data indexed by line number - to be imported to Pandas dataframe
    header_dat = {}
    
    # create empty array to hold line indices
    line_counter = []
    
    # generate a list of header lines - contains all data we need to write to project set up database
    line_list = []
    
    # open file passed to function
    o_file = open(file_name, encoding='utf-8')
    
    # start line counter
    counter = 0
    
    # start end of file 
    eof = 0
    line_counter.append(counter)
    
    # read first line in file
    line = o_file.readline()[:-1]
    
    # append the current line of header data to the line list
    line_list.append(line)
    
    with o_file as f:
        for line in f:
            if "** Data Segment **" in line:
                # if this current line signifies the start of the data stream
                counter = counter + 1
                # the data starts 5 rows down from this
                dataRow = counter + 5
                # break the loop, we have reached our stop point
                break

            else:
                # if we are still reading header data increase the line counter by 1
                counter = counter + 1
                # append the line counter to the count array
                line_counter.append(counter)
                # append line of data to the data array
                line_list.append(line)
    
    # add count array to dictionary with field name 'idx' as key
    header_dat['idx'] = line_counter
    
    # add data line array to dictionary with field name 'line' as key
    header_dat['line'] = line_list
    
    # create pandas dataframe of header data indexed by row number
    headerDF = pd.DataFrame.from_dict(header_dat)
    headerDF.set_index('idx',inplace = True)
    
    # create a variable for header type because apparently Lotek makes its mind up
    header_type = 'standard'

    # gotta love multiple header types 
    for row in headerDF.iterrows():
        if row[1][0] == 'Environment History:':
            header_type = 'alternate'
            break
    del row
        
    # find scan time
    # for every header data row
    for row in headerDF.iterrows():   
        if 'scan time' in row[1][0] or 'Scan Time' in row[1][0]:
            # get the number value from the row
            scanTimeStr = row[1][0][-7:-1]
            # split the string
            scanTimeSplit = scanTimeStr.split(':')
            # convert the scan time string to float
            scanTime = float(scanTimeSplit[1])
            # stop that loop, we done
            break
    del row
    
    # find number of channels and create channel dictionary
    # create empty array of channel ID's
    scanChan = []
    
    # create empty channel ID: frequency dictionary
    channelDict = {}
    
    # create counter
    counter = 0
    
    # create row iterator
    rows = headerDF.iterrows()
    
    for row in rows:
        # if the first 18 characters say Active scan_table
        if 'Active scan_table:' in row[1][0]:
            # channel dictionary data starts two rows from here
            idx0 = counter + 2
            # while the next row isn't empty
            while next(rows)[1][0] != '\n':
                # increase the counter
                counter = counter + 1
            # get index of last data row
            idx1 = counter + 1
            break
        else:
            counter = counter + 1
    del row, rows

    # extract channel dictionary data using rows identified earlier
    channelDat = headerDF.iloc[idx0:idx1]
    for row in channelDat.iterrows():
        # extract that channel ID from the data row and append to array
        dat = row[1][0]
        channel = int(dat[0:4])
        frequency = dat[10:17]
        channelDict[channel] = frequency
        scanChan.append(channel)
    channels = len(scanChan)

    def id_to_freq(row,channelDict):
        channel = row['ChannelID']
        if np.int(channel) in channelDict:
            return channelDict[np.int(channel)]
        else:
            return '888'
        
    if header_type == 'alternate':

        # with our data row, extract information using pandas fwf import procedure
        telem_dat = pd.read_fwf(os.path.join(file_name),
                               colspecs = [(0,5),(5,14),(14,20),(20,26),(26,30),(30,35)],
                               names = ['DayNumber','Time','ChannelID','Power','Antenna','TagID'],
                               skiprows = dataRow)
        
        # remove last two rows, Lotek adds garbage at the end
        telem_dat = telem_dat.iloc[:-2]
        telem_dat.dropna(inplace = True)
        
        if len(telem_dat) > 0:
            for ant in ant_to_rec_dict:
                site = ant_to_rec_dict[ant]
                telem_dat_sub = telem_dat[telem_dat.Antenna == ant]
                
                # calculate frequency with channel ID
                telem_dat_sub['Frequency'] = telem_dat_sub['ChannelID'].map(channelDict) #(id_to_freq, axis = 1, args = (channelDict,))
        
                # remove some extraneous data
                telem_dat_sub = telem_dat_sub[telem_dat_sub.Frequency != '888']
                telem_dat_sub = telem_dat_sub[telem_dat_sub.TagID != 999]
                
                # create freq code column
                telem_dat_sub['freq_code'] = telem_dat_sub['Frequency'].astype(str) + ' ' + telem_dat_sub['TagID'].astype(int).astype(str)
                
                # deal with that pesky number date and clean up
                telem_dat_sub['day0'] = np.repeat(pd.to_datetime("1900-01-01"),len(telem_dat_sub))
                telem_dat_sub['Date'] = telem_dat_sub['day0'] + pd.to_timedelta(telem_dat_sub['DayNumber'].astype(int), unit='d')
                telem_dat_sub['Date'] = telem_dat_sub.Date.astype('str')
                telem_dat_sub['time_stamp'] = pd.to_datetime(telem_dat_sub['Date'] + ' ' + telem_dat_sub['Time'])
                telem_dat_sub.drop(['day0','DayNumber'],axis = 1, inplace = True)
                
                # calculate unix epoch
                telem_dat['epoch'] = np.round((telem_dat.time_stamp - pd.Timestamp("1970-01-01")) / pd.Timedelta('1s'),6)
                
                # clean up some more
                telem_dat_sub.drop (['Date','Time','Frequency','TagID','ChannelID','Antenna'],axis = 1, inplace = True)
                
                # add file name to data
                #telemDat_sub['fileName'] = np.repeat(rxfile,len(telemDat_sub))
                
                # add recID
                telem_dat_sub['rec_id'] = np.repeat(site,len(telem_dat_sub))
                
                # calculate noise ratio
                telem_dat_sub = predictors.noise_ratio(600,
                                           telem_dat_sub.freq_code.values,
                                           telem_dat_sub.epoch.values,
                                           study_tags)
                
                # add in scan time, channels, rec type
                telem_dat_sub['scan_time'] = np.repeat(scan_time,len(telem_dat_sub))
                telem_dat_sub['channels'] = np.repeat(channels,len(telem_dat_sub))
                telem_dat_sub['rec_type'] = np.repeat(rec_type,len(telem_dat_sub))
                
                # make multiindex for fast indexing
                # tuples = zip(telem_dat_sub.FreqCode.values,telem_dat_sub.recID.values,telem_dat_sub.Epoch.values)
                # index = pd.MultiIndex.from_tuples(tuples, names=['FreqCode', 'recID','epoch'])
                # telem_dat_sub.set_index(index,inplace = True,drop = False)
                telem_dat_sub.reset_index(inplace = True)
                telem_dat_sub['noise_ratio'] = telem_dat_sub.noise_ratio.values.astype(np.float32)

                # write to SQL
                
                telem_dat_sub = telem_dat_sub.astype({'power':'float32',
                                                      'freq_code':'object',
                                                      'time_stamp':'datetime64[ns]',
                                                      'scan_time':'float32',
                                                      'channels':'int32',
                                                      'rec_type':'object',
                                                      'epoch':'float32',
                                                      'noise_ratio':'float32',
                                                      'rec_id':'object'})
                
                with pd.HDFStore(db_dir, mode='a') as store:
                    store.append(key = 'raw_data',
                                 value = telem_dat_sub, 
                                 format = 'table', 
                                 index = False,
                                 min_itemsize = {'freq_code':20,
                                                 'rec_type':20,
                                                 'rec_id':20},
                                 append = True, 
                                 chunksize = 1000000,
                                 data_columns = True)    
    else:
        telem_dat = pd.read_fwf(file_name,
                               colspecs = [(0,9),(9,19),(19,29),(29,36),(36,44),(44,52)],
                               names = ['Date','Time','ChannelID','TagID','Antenna','Power'],
                               skiprows = dataRow)
        
        # remove last two rows, Lotek adds garbage at the end
        telem_dat = telem_dat.iloc[:-2]
        telem_dat.dropna(inplace = True)
        
        if len(telem_dat) > 0:
            for ant in ant_to_rec_dict:
                site = ant_to_rec_dict[ant]
                telem_dat_sub = telem_dat[telem_dat.Antenna == ant]
                
                # calculate frequency with channel ID
                telem_dat_sub['Frequency'] = telem_dat_sub['ChannelID'].map(channelDict)
                
                # remove extra data
                telem_dat_sub = telem_dat_sub[telem_dat_sub.Frequency != '888']
                telem_dat_sub = telem_dat_sub[telem_dat_sub.TagID != 999]
                
                
                # calculate freqcode
                telem_dat_sub['freq_code'] = telem_dat_sub['Frequency'].astype(str) + ' ' + telem_dat_sub['TagID'].astype(int).astype(str)
                
                # get timestamp
                telem_dat_sub['time_stamp'] = pd.to_datetime(telem_dat_sub['Date'] + ' ' + telem_dat_sub['Time'])# create timestamp field from date and time and apply to index
                
                # calculate UNIX epoch 
                telem_dat_sub['epoch'] = np.round((telem_dat_sub.time_stamp - pd.Timestamp("1970-01-01")) / pd.Timedelta('1s'),6)
                
                # calculate noise ratio
                telem_dat_sub = predictors.noise_ratio(5.0,
                                           telem_dat_sub.freq_code.values,
                                           telem_dat_sub.epoch.values,
                                           study_tags)
                
                # clean up some more
                telem_dat_sub.drop (['Date','Time','Frequency','TagID','ChannelID','Antenna'],axis = 1, inplace = True)
                
                # add in scan time, channels, rec type
                telem_dat_sub['scan_time'] = np.repeat(scan_time,len(telem_dat_sub))
                telem_dat_sub['channels'] = np.repeat(channels,len(telem_dat_sub))
                telem_dat_sub['rec_type'] = np.repeat(rec_type,len(telem_dat_sub))
                telem_dat_sub['rec_id'] = np.repeat(site,len(telem_dat_sub))
                
                # make multiindex for fast indexing
                # tuples = zip(telem_dat_sub.FreqCode.values,telem_dat_sub.recID.values,telem_dat_sub.epoch.values)
                # index = pd.MultiIndex.from_tuples(tuples, names=['FreqCode', 'recID','epoch'])
                # telem_dat_sub.set_index(index,inplace = True,drop = False)
                telem_dat_sub.reset_index(inplace = True)
                telem_dat_sub['noise_ratio'] = telem_dat.noise_ratio.values.astype(np.float32)

                telem_dat_sub = telem_dat_sub.astype({'power':'float32',
                                                      'freq_code':'object',
                                                      'time_stamp':'datetime64[ns]',
                                                      'scan_time':'float32',
                                                      'channels':'int32',
                                                      'rec_type':'object',
                                                      'epoch':'float32',
                                                      'noise_ratio':'float32',
                                                      'rec_id':'object'})

                # write to SQL
                with pd.HDFStore(db_dir, mode='a') as store:
                    store.append(key = 'raw_data',
                                 value = telem_dat_sub, 
                                 format = 'table', 
                                 index = False,
                                 min_itemsize = {'freq_code':20,
                                                 'rec_type':20,
                                                 'rec_id':20},
                                 append = True, 
                                 chunksize = 1000000)
    
    
    
    
    
    
    


# Import the submodules first that have no dependencies
from .predictors import *
from .overlap_removal import *
from .table_merge import *
from .formatter import *

# Import the naive_bayes submodule next
from .naive_bayes import *

# Import the overlap removal submodule
from .overlap_removal import *

# Import the parsers submodule, which depends on predictor
from .parsers import *

# Finally, import the radio_project class, which depends on parsers, predictor, and naive_bayes

from .radio_project import *






# -*- coding: utf-8 -*-
'''
Module contains all of the methods and classes required to identify and remove
overlapping detections from radio telemetry data.
'''

# import modules required for function dependencies
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit, minimize
from scipy.interpolate import UnivariateSpline
import matplotlib.pyplot as plt
import networkx as nx
from matplotlib import rcParams

font = {'family': 'serif','size': 6}
rcParams['font.size'] = 6
rcParams['font.family'] = 'serif'

class bout():
    '''Python class object to delineate when bouts occur at receiver.'''
    def __init__ (self, radio_project, node, lag_window, time_limit):
        self.lag_window = lag_window
        self.time_limit = time_limit
        self.db = radio_project.db

        # get the receivers associated with this particular network node
        recs = radio_project.receivers[radio_project.receivers.node == node]
        self.receivers = recs.index  # get the unique receivers associated with this node
        self.data = pd.DataFrame(columns = ['freq_code','epoch','rec_id'])            # set up an empty data frame

        # for every receiver
        for i in self.receivers:
            # get this receivers data from the classified key
            rec_dat = pd.read_hdf(self.db,
                                  'classified',
                                  where = 'rec_id = %s'%(i))
            rec_dat = rec_dat[rec_dat.iter == rec_dat.iter.max()]
            rec_dat = rec_dat[rec_dat.test == 1]
            rec_dat = rec_dat[['freq_code','epoch','rec_id']]
            rec_dat = rec_dat.astype({'freq_code':'object',
                            'epoch':'float32',
                            'rec_id':'object'})
            
            self.data = self.data.append(rec_dat)

        # clean up and bin the lengths
        self.data.drop_duplicates(keep = 'first', inplace = True)
        self.data.sort_values(by = ['freq_code','epoch'], inplace = True)
        self.data['det_lag'] = self.data.groupby('freq_code')['epoch'].diff().abs() // lag_window * lag_window
        self.data.dropna(axis = 0, inplace = True)   # drop Nan from the data
        self.node = node
        self.fishes = self.data.freq_code.unique()
        
    def prompt_for_params(self, model_type):
        if model_type == 'two_process':
            print("Please enter the values for the initial quantity (y0), the quantity at time t (yt), and the time t.")
            y0 = float(input("Enter the initial quantity (y0): "))
            yt = float(input("Enter the quantity at time t (yt): "))
            t = float(input("Enter the time at which yt is observed (t): "))
        
            # Calculate decay rate b1 using the provided y0, yt, and t
            b1 = -np.log(yt / y0) / t
        
            # Assume that the decay rate b2 after the knot is the same as b1 before the knot
            # This is a simplifying assumption; you may want to calculate b2 differently based on additional data or domain knowledge
            b2 = b1
        
            # For the two-process model, we'll assume a1 is the initial quantity y0
            a1 = y0
        
            # We'll calculate a2 such that the function is continuous at the knot
            # This means solving the equation a1 * exp(-b1 * k) = a2 * exp(-b2 * k)
            # Since we've assumed b1 = b2, this simplifies to a2 = a1 * exp(-b1 * k)
            a2 = a1 * np.exp(-b1 * t)
        
            return [a1, b1, a2, b2, t]
        
        else:
            print ("Sorry, we don't yet support that model type")
        # get lag frequencies
        lags = np.arange(0,self.time_limit,2)
        freqs, bins = np.histogram(np.sort(self.data.det_lag),lags)
        bins = bins[np.where(freqs > 0)]
        freqs = freqs[np.where(freqs > 0)]
        log_freqs = np.log(freqs)
        
        # Plot the raw data
        plt.scatter(bins, log_freqs, label='Data')
        plt.xlabel('Lag')
        plt.ylabel('Lag Frequency')
        plt.title('Raw Data for Two-Process Model')
        plt.legend()
        plt.show()
    
        # Prompt the user for initial parameters
        initial_guess = self.prompt_for_params(model_type = 'two_process')
    
        # Perform the curve fitting
        try:
            params, params_covariance = curve_fit(self.two_process_model, 
                                                  bins, 
                                                  log_freqs, 
                                                  p0=initial_guess)
            
            # Plot the fitted function
            plt.plot(bins, self.two_process_model(bins, *params), 
                     label='Fitted function',
                     color='red')
            
            plt.scatter(bins, log_freqs, label='Data')
            plt.legend()
            plt.xlabel('x')
            plt.ylabel('y')
            plt.title('Fitted Two-Process Model')
            plt.show()
            
            # Return the fitted parameters
            return params
        except RuntimeError as e:
            print(f"An error occurred during fitting: {e}")
            return None
        
    def find_knot(self, initial_knot_guess):
        # get lag frequencies
        lags = np.arange(0, self.time_limit, 2)
        freqs, bins = np.histogram(np.sort(self.data.det_lag), lags)
        bins = bins[:-1][freqs > 0]  # Ensure the bins array is the right length
        freqs = freqs[freqs > 0]
        log_freqs = np.log(freqs)
        
        # Define a two-segment exponential decay function
        def two_exp_decay(x, a1, b1, a2, b2, k):
            condlist = [x < k, x >= k]
            funclist = [
                lambda x: a1 * np.exp(-b1 * x),
                lambda x: a2 * np.exp(-b2 * (x - k))
            ]
            return np.piecewise(x, condlist, funclist)


        # Objective function for two-segment model
        def objective_function(knot, bins, log_freqs):
            # Fit the model without bounds on the knot
            try:
                params, _ = curve_fit(lambda x, a1, b1, a2, b2: two_exp_decay(x, a1, b1, a2, b2, knot),
                                      bins, 
                                      log_freqs, 
                                      p0=[log_freqs[0], 0.001, 
                                          log_freqs[0], 0.001])
                y_fit = two_exp_decay(bins, *params, knot)
                error = np.sum((log_freqs - y_fit) ** 2)
                return error
            except RuntimeError:
                return np.inf

        # Use minimize to find the optimal knot
        result = minimize(
            fun=objective_function,
            x0=[initial_knot_guess],
            args=(bins, log_freqs),
            bounds=[(bins.min(), bins.max())]
        )

        # Check if the optimization was successful and extract the results
        if result.success:
            optimized_knot = result.x[0]

            # Refit with the optimized knot to get all parameters
            p0 = [log_freqs[0], 0.001, 
                  log_freqs[0], 0.001,
                  optimized_knot]
            
            bounds_lower = [0, 0, 
                            0, 0, 
                            bins.min()]
            
            bounds_upper = [np.inf, np.inf, 
                            np.inf, np.inf, 
                            bins.max()]

            optimized_params, _ = curve_fit(
                two_exp_decay,
                bins,
                log_freqs,
                p0=p0,
                bounds=(bounds_lower, bounds_upper)
            )

            # Visualization of the final fit with the estimated knot
            plt.figure(figsize=(12, 6))
            
            plt.scatter(bins, 
                        log_freqs, 
                        label='Data', 
                        alpha=0.6)
            
            x_range = np.linspace(bins.min(), bins.max(), 1000)
            
            plt.plot(x_range, 
                     two_exp_decay(x_range, *optimized_params),
                     label='Fitted function', 
                     color='red')
            
            plt.axvline(x=optimized_knot, color='orange', linestyle='--')#, label=f'Knot at x={optimized_knot:.2f}')
            plt.title('Fitted Two-Process Model')
            plt.xlabel('Lag Time')
            plt.ylabel('Frequency')
            plt.legend()
            plt.show()
            
            return optimized_params[-1]
        
        else:
            print("Optimization failed:", result.message)
            return None
    
    def find_knots(self, initial_knot_guesses):
        # get lag frequencies
        lags = np.arange(0, self.time_limit, 2)
        freqs, bins = np.histogram(np.sort(self.data.det_lag), lags)
        bins = bins[:-1][freqs > 0]  # Ensure the bins array is the right length
        freqs = freqs[freqs > 0]
        log_freqs = np.log(freqs)
        # Assuming initial_knot_guesses is a list of two knot positions
        # This method should fit a three-process model

        # Define bounds for parameters outside of objective_function
        bounds_lower = [0, 0, 0, 0, 0, 0, bins.min(), initial_knot_guesses[0]]
        bounds_upper = [np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, initial_knot_guesses[1], bins.max()]
    
        # Define a three-segment exponential decay function
        #TODO - is the math correct?
        def three_exp_decay(x, a1, b1, a2, b2, a3, b3, k1, k2):
            condlist = [x < k1, (k1 <= x) & (x < k2), x >= k2]
            funclist = [
                lambda x: a1 * np.exp(-b1 * x),
                lambda x: a2 * np.exp(-b2 * (x - k1)),
                lambda x: a3 * np.exp(-b3 * (x - k2))
            ]
            return np.piecewise(x, condlist, funclist)
    
        # Objective function for three-segment model
        def objective_function(knots, bins, log_freqs):
            # Unpack knots
            k1, k2 = knots
            # Initial parameter guesses (3 amplitudes, 3 decay rates)
            p0 = [log_freqs[0], 0.001, log_freqs[0], 0.001, log_freqs[0], 0.001, k1, k2]

            # Fit the three-segment model
            params, _ = curve_fit(three_exp_decay,
                                  bins, 
                                  log_freqs,
                                  p0=p0, 
                                  bounds=(bounds_lower, bounds_upper))            # Calculate the errors
            #TODO - AIC or BIC possible expansion?
            y_fit = three_exp_decay(bins, *params)
            error = np.sum((log_freqs - y_fit) ** 2)
            return error
    
        # Perform the optimization with the initial guesses
        result = minimize(
            fun=objective_function,
            x0=initial_knot_guesses,  # Initial guesses for the knot positions
            args=(bins, log_freqs),
            bounds=[(bins.min(), initial_knot_guesses[1]),  # Ensure k1 < k2
                    (initial_knot_guesses[0], bins.max())]
        )
    
        # Check if the optimization was successful and extract the results
        if result.success:
            optimized_knots = result.x  # These should be the optimized knots
    
            # Now we refit the model with the optimized knots to get all the parameters
            p0 = [log_freqs[0], 0.001, 
                  log_freqs[0], 0.001, 
                  log_freqs[0], 0.001, 
                  optimized_knots[0], optimized_knots[1]]
            
            bounds_lower = [0, 0, 
                            0, 0, 
                            0, 0, 
                            bins.min(), optimized_knots[0]]
            
            bounds_upper = [np.inf, np.inf, 
                            np.inf, np.inf, 
                            np.inf, np.inf, 
                            optimized_knots[1], bins.max()]
    
            optimized_params, _ = curve_fit(
                three_exp_decay,
                bins,
                log_freqs,
                p0=p0,
                bounds=(bounds_lower, bounds_upper)
            )           

            # Visualization of the final fit with the estimated knots
            plt.figure(figsize=(12, 6))
            plt.scatter(bins, log_freqs, label='Data', alpha=0.6)

            # Create a range of x values for plotting the fitted function
            x_range = np.linspace(bins.min(), bins.max(), 1000)
            plt.plot(x_range, 
                     three_exp_decay(x_range, *optimized_params), 
                     label='Fitted function', color='red')

            plt.axvline(x=optimized_knots[0], color='orange', linestyle='--')
            plt.axvline(x=optimized_knots[1], color='green', linestyle='--')

            plt.title('Fitted Three-Process Model')
            plt.xlabel('Lag Time')
            plt.ylabel('Frequency')
            plt.legend()
            plt.show()
        else:
            print("Optimization failed:", result.message)
            return None

        return optimized_params[-1]
    
    def fit_processes(self):
        # Step 1: Plot bins vs log frequencies
        lags = np.arange(0, self.time_limit, 2)
        freqs, bins = np.histogram(np.sort(self.data.det_lag), lags)
        bins = bins[:-1][freqs > 0]  # Ensure the bins array is the right length
        freqs = freqs[freqs > 0]
        log_freqs = np.log(freqs)
        
        plt.figure(figsize=(12, 6))
        plt.scatter(bins, log_freqs, label='Log of Frequencies', alpha=0.6)
        plt.title('Initial Data Plot')
        plt.xlabel('Bins')
        plt.ylabel('Log Frequencies')
        plt.legend()
        plt.show()
    
        # Step 2: Ask user for initial knots
        num_knots = int(input("Enter the number of knots (1 for two-process, 2 for three-process): "))
        initial_knots = []
        for i in range(num_knots):
            knot = float(input(f"Enter initial guess for knot {i+1}: "))
            initial_knots.append(knot)
    
        # Step 3 & 4: Determine the number of processes and fit accordingly
        if num_knots == 1:
            # Fit a two-process model
            self.initial_knot_guess = initial_knots[0]
            self.find_knot(self.initial_knot_guess)
        elif num_knots == 2:
            # Fit a three-process model (you will need to implement this method)
            self.initial_knot_guesses = initial_knots
            self.find_knots(self.initial_knot_guesses)
        else:
            print("Invalid number of knots. Please enter 1 or 2.")
            
        # Fit the model based on the number of knots
        optimized_knots = None
        if num_knots == 1:
            # Fit a two-process model
            optimized_knots = self.find_knot(initial_knots[0])
        elif num_knots == 2:
            # Fit a three-process model
            optimized_knots = self.find_knots(initial_knots)
        else:
            print("Invalid number of knots. Please enter 1 or 2.")
        
        # Return the optimized knot(s)
        return optimized_knots
      
    def presence(self, threshold):
        '''Function takes the break point between a continuous presence and new presence,
        enumerates the presence number at a receiver and writes the data to the
        analysis database.'''
        fishes = self.data.freq_code.unique()
        
        for fish in fishes:
            fish_dat = self.data[self.data.freq_code == fish]
        
            # Vectorized classification
            classifications = np.where(fish_dat.det_lag <= threshold, 'within_bout', 'start_new_bout')
        
            # Generating bout numbers
            # Increment bout number each time a new bout starts
            bout_changes = np.where(classifications == 'start_new_bout', 1, 0)
            bout_no = np.cumsum(bout_changes)
        
            # Assigning classifications and bout numbers to the dataframe
            fish_dat['class'] = classifications
            fish_dat['bout_no'] = bout_no 
            
            fish_dat = fish_dat.astype({'freq_code': 'object',
                                        'epoch': 'float32',
                                        'rec_id': 'object',
                                        'class': 'object',
                                        'bout_no':'int32',
                                        'det_lag':'int32'})
    
            # append to hdf5
            with pd.HDFStore(self.db, mode='a') as store:
                store.append(key = 'presence',
                             value = fish_dat, 
                             format = 'table', 
                             index = False,
                             min_itemsize = {'freq_code':20,
                                             'rec_id':20,
                                             'class':20},
                             append = True, 
                             data_columns = True,
                             chunksize = 1000000)
        
            print ('bouts classified for fish %s'%(fish))
            

class overlap_reduction():
    '''Python class  to reduce redundant dections at overlappin receivers.
    More often than not, a large aerial yagi will be placed adjacent to a dipole.
    The yagi has a much larger detection area and will pick up fish in either detection
    zone.  The dipole is limited in its coverage area, therefore if a fish is
    currently present at a dipole and is also detected at the yagi, we can safely
    assume that the detection at the Yagi are overlapping and we can place the fish
    at the dipole antenna.  By identifying and removing these overlapping detections
    we remove bias in time-to-event modeling when we want to understand movement
    from detection areas with limited coverage to those areas with larger aerial
    coverages.

    This class object contains a series of methods to identify overlapping detections
    and import a table for joining into the project database.'''

    def __init__(self, nodes, edges, radio_project):
        '''The initialization module imports data and creates a networkx graph object.

        The end user supplies a list of nodes, and a list of edges with instructions
        on how to connect them and the function does the rest.  NO knowlege of networkx
        is required.

        The nodes and edge relationships should start with the outermost nodes and
        eventually end with the inner most node/receiver combinations.

        Nodes must be a list of nodes and edges must be a list of tuples.
        Edge example: [(1,2),(2,3)],
        Edges always in format of [(from,to)] or [(outer,inner)] or [(parent,child)]'''
        self.db = radio_project.db

        # Step 1, create a directed graph from list of edges
        self.G = nx.DiGraph()
        self.G.add_edges_from(edges)
        
        # Step 2, import data and create a dictionary of node dataframes
        self.node_pres_dict = dict()
        self.node_recap_dict = dict()
        
        for i in nodes:
            #import data and add to node dict
            node_recs = radio_project.receivers[radio_project.receivers.node == i]
            node_recs = node_recs.index         # get the unique receivers associated with this node
            pres_data = pd.DataFrame(columns = ['freq_code','epoch','node','rec_id','presence'])        # set up an empty data frame
            recap_data = pd.DataFrame(columns = ['freq_code','epoch','node','rec_id'])
            
            for j in node_recs:
                # get presence data and final classifications for this receiver
                presence_dat = pd.read_hdf(radio_project.db,'presence', where = 'rec_id = %s'%(j))
                presence_dat['node'] = np.repeat(i,len(presence_dat))
                class_dat = pd.read_hdf(radio_project.db,'classified', where = 'rec_id = %s'%(j))
                class_dat = class_dat[class_dat.iter == class_dat.iter.max()]
                class_dat = class_dat[class_dat.test == 1]
                class_dat = class_dat[['freq_code', 'epoch', 'rec_id']]
                class_dat['node'] = np.repeat(i,len(class_dat))
                # append to node specific dataframe
                pres_data = pres_data.append(presence_dat)
                recap_data = recap_data.append(class_dat)

            # now that we have data, we need to summarize it, use group by to get min ans max epoch by freq code, recID and presence_number
            dat = pres_data.groupby(['freq_code','bout_no','node','rec_id'])['epoch'].agg(['min','max'])
            dat.reset_index(inplace = True, drop = False)
            dat.rename(columns = {'min':'min_epoch','max':'max_epoch'},inplace = True)
            self.node_pres_dict[i] = dat
            self.node_recap_dict[i] = recap_data
            
            # clean up
            del pres_data, recap_data, dat, presence_dat, class_dat
            print ("Completed data management process for node %s"%(i))

        # visualize the graph
        shells = []
        for n in list(self.G.nodes()):
            successors = list(self.G.succ[n].keys())
            shells.append(successors)

        fig, ax = plt.subplots(1, 1, figsize=(4, 4));
        pos= nx.circular_layout(self.G)
        nx.draw_networkx_nodes(self.G,pos,list(self.G.nodes()),node_color = 'r',node_size = 400)
        nx.draw_networkx_edges(self.G,pos,list(self.G.edges()),edge_color = 'k')
        nx.draw_networkx_labels(self.G,pos,font_size=8)
        plt.axis('off')
        plt.show()

    def nested_doll(self):
        '''Function iterates through matching recap data from successors to see if
        current recapture row at predeccesor overlaps with successor presence.'''
        # create function that we can vectorize over list of epochs (i)
        def overlap_check(i, min_epoch, max_epoch):
            return np.logical_and(min_epoch >= i, max_epoch < i).any()
        for i in self.node_recap_dict:
            fishes = self.node_recap_dict[i].freq_code.unique()

            for j in fishes:
                children = self.G.succ[i]
                fish_dat = self.node_recap_dict[i][self.node_recap_dict[i].freq_code == j]
                fish_dat['overlapping'] = np.zeros(len(fish_dat))
                fish_dat['parent'] = np.repeat('',len(fish_dat))
                fish_dat.set_index('epoch', inplace = True, drop = False)
                if len(children) > 0:                                            # if there is no child node, who cares? there is no overlapping detections, we are at the middle doll
                    for k in children:
                        child_dat = self.node_pres_dict[i][self.node_pres_dict[i].freq_code == j]
                        if len(child_dat) > 0:
                            for l in child_dat.rec_id.unique():
                                while l != i:
                                    rec_dat = child_dat[child_dat.rec_id == l]
                                    min_epochs = child_dat.min_epoch.values
                                    max_epochs = child_dat.max_epoch.values
                                    for m in fish_dat.epoch.values:                          # for every row in the fish data
                                        if np.logical_and(min_epochs <= m, max_epochs >m).any(): # if the current epoch is within a presence at a child receiver
                                            print ("Overlap Found, at %s fish %s was recaptured at both %s and %s"%(m,j,i,l))
                                            fish_dat.at[m,'overlapping'] = 1
                                            fish_dat.at[m,'parent'] = i
                                            
                fish_dat.reset_index(inplace = True, drop = True)
                
                fish_dat = fish_dat.astype({'freq_code': 'object',
                                            'epoch': 'int32',
                                            'rec_id': 'object',
                                            'node': 'object',
                                            'overlapping':'int32',
                                            'parent':'object'})
                
                # append to hdf5
                with pd.HDFStore(self.db, mode='a') as store:
                    store.append(key = 'overlapping',
                                 value = fish_dat, 
                                 format = 'table', 
                                 index = False,
                                 min_itemsize = {'freq_code':20,
                                                 'rec_id':20,
                                                 'parent':20},
                                 append = True, 
                                 data_columns = True,
                                 chunksize = 1000000)


# -*- coding: utf-8 -*-
"""
Modules contains all of the functions and classes required to format radio
telemetry data for statistical testing.

currently we can set up models for Cormack Jolly Seber mark recapture, Time-to-
Event modeling, and Live Recapture Dead Recovery Mark Recapture.
"""

# import modules required for function dependencies
import numpy as np
import pandas as pd
import os
import sqlite3
import datetime
from matplotlib import rcParams
from scipy import interpolate
from pymast.radio_project import radio_project


font = {'family': 'serif','size': 6}
rcParams['font.size'] = 6
rcParams['font.family'] = 'serif'

class cjs_data_prep():
    '''Class creates input files for Cormack Jolly Seber modeling in MARK'''
    def __init__(self, 
                 receiver_to_recap, 
                 project, 
                 input_type = 'query',
                 species = None, 
                 rel_loc = None, 
                 cap_loc = None, 
                 initial_recap_release = False):
        
        # Import Data From Project HDF
        self.rel_loc = rel_loc
        self.cap_loc = cap_loc

        # get recaptures, but first build a query
        query_parts = []
        for key in receiver_to_recap:
            query_parts.append(f"rec_id == '{key}'")   
        qry = " & ".join(query_parts)

        self.recap_data = pd.read_hdf(project.db,
                             'recaptures')#, 
                             #where = qry)
        
        #self.recap_data.set_index('freq_code', inplace = True)
        project.tags.reset_index('freq_code', inplace = True)
        self.recap_data = pd.merge(self.recap_data, 
                                   project.tags,
                                   left_on = 'freq_code', 
                                   right_on = 'freq_code',
                                   how = 'left')
        self.recap_data.reset_index(drop = False, inplace = True)
        #project.tags.reset_index(drop = False, inplace = True)
        
        # filter out tag data we don't want mucking up our staistical model
        if species != None:
            self.recap_data = self.recap_data[self.recap_data.Species == species]
        if rel_loc != None:
            self.recap_data = self.recap_data[self.recap_data.RelLoc == rel_loc]
        if cap_loc != None:
            self.recap_data = self.recap_data[self.recap_data.CapLoc == cap_loc]

        # map recap occasions
        self.recap_data['recap_occasion'] = self.recap_data['rec_id'].map(receiver_to_recap)

        if initial_recap_release == False:
            '''When initial recap release is true, the first recapture occasion
            in the rec_list query becomes our release location - this is helpful
            when modeling downstream migratory behavior of anadromous fish like
            American Shad.  We don't care about what happens on their upstream
            migration, we only care about when they turn around and come back down.

            When this argument is false, we are modeling fish from their initial
            release time and we add in R00 as our release location, otherwise R00
            should appear on the input receiver to recapture event dictionary.
            '''

            print ("Adding release time for this fish")
            rel_dat = project.tags
            if self.rel_loc is not None:
                rel_dat = rel_dat[rel_dat.rel_loc == rel_loc]
            if self.cap_loc is not None:
                rel_dat = rel_dat[rel_dat.cap_loc == cap_loc]
            rel_dat['rel_date'] = pd.to_datetime(rel_dat.rel_date)
            rel_dat['epoch'] = (rel_dat['rel_date'] - datetime.datetime(1970,1,1)).dt.total_seconds()
            rel_dat.rename(columns = {'rel_date':'time_stamp'}, inplace = True)
            rel_dat['recap_occasion'] = np.repeat('R00',len(rel_dat))
            rel_dat['overlapping'] = np.zeros(len(rel_dat))
            self.recap_data = pd.concat([self.recap_data,rel_dat],axis = 0)

        else:
            print ("Starting Initial Recap Release Procedure")
            # Identify first recapture times
            start_times = self.recap_data[self.recap_data.recap_occasion == "R00"].groupby(['freq_code'])['epoch'].min().to_frame()
            start_times.reset_index(drop = False, inplace = True)
            start_times.set_index('freq_code',inplace = True)
            start_times.rename(columns = {'epoch':'first_recapture'},inplace = True)

            # make sure fish start from the first recapture occassion and that there are no recaps before release
            for fish in self.recap_data.freq_code.unique():

                if fish not in start_times.index.values:
                    # fish never made it to the initial state
                    self.recap_data.drop(self.recap_data[self.recap_data.freq_code == fish].index, inplace = True)
                else:
                    # fish arrived at the initial state but their may be recaptures before arrival at initial state
                    t = start_times.at[fish,'first_recapture']
                    self.recap_data.drop(self.recap_data[(self.recap_data.freq_code == fish) & (self.recap_data.epoch < t)].index, inplace = True)

        print (self.recap_data.head())

    def input_file(self,model_name, output_ws):
        #Step 1: Create cross tabulated data frame with FreqCode as row index and recap occasion as column
        cross_tab = pd.pivot_table(self.recap_data, values = 'epoch', index = 'freq_code', columns = 'recap_occasion', aggfunc = 'min')

        #Step 2: Fill in those nan values with 0, we can't perform logic on nothing!'''
        cross_tab.fillna(value = 0, inplace = True)

        #Step 3: Map a simply if else statement, if value > 0,1 else 0'''
        cross_tab = cross_tab.applymap(lambda x:1 if x > 0 else 0)

        # put together the input string
        inp = "/* " + cross_tab.index.astype(str) + " */  "
        for i in np.arange(0,len(cross_tab.columns),1):
            inp = inp + cross_tab.iloc[:,i].astype(str)
        inp = inp + "     1;"

        self.inp = inp
        self.cross = cross_tab
        # Check your work
        print (cross_tab.head(100))
        cross_tab.to_csv(os.path.join(output_ws,'%s_cjs.csv'%(model_name)))

class lrdr_data_prep():
    '''Class creates input files for Live Recapture Dead Recovery modeling in MARK'''
    def __init__(self,receiver_to_recap,mobile_to_recap, dbDir, input_type = 'query', rel_loc = None, cap_loc = None, initial_recap_release = False, time_limit = None):

        # Import Data using sql
        '''the default input type for this function is query, however'''
        print ("Starting extraction of recapture data related to the recievers chosen")
        conn = sqlite3.connect(dbDir)
        c = conn.cursor()
        self.rel_loc = rel_loc
        self.cap_loc = cap_loc
        # get live recapture data
        receiver_list = list(receiver_to_recap.keys())
        sql = 'SELECT tblRecaptures.FreqCode, Epoch, timeStamp, tblRecaptures.recID, TagType, overlapping, RelLoc, CapLoc, test FROM tblRecaptures LEFT JOIN tblMasterReceiver ON tblRecaptures.recID = tblMasterReceiver.recID LEFT JOIN tblMasterTag ON tblRecaptures.FreqCode = tblMasterTag.FreqCode WHERE tblRecaptures.recID = "%s"'%(receiver_list[0])
        for i in receiver_list[1:]:
            sql = sql + ' OR tblRecaptures.recID = "%s"'%(i)
        self.live_recap_data = pd.read_sql_query(sql,con = conn,coerce_float  = True)
        #self.live_recap_data['timeStamp'] = self.live_recap_data['timeStamp'].to_datetime()
        self.live_recap_data = self.live_recap_data[self.live_recap_data.TagType == 'Study']
        self.live_recap_data = self.live_recap_data[self.live_recap_data.overlapping == 0.0]
        self.live_recap_data = self.live_recap_data[self.live_recap_data.test == 1.0]
        if rel_loc != None:
            self.live_recap_data = self.live_recap_data[self.live_recap_data.RelLoc == rel_loc]
        if cap_loc != None:
            self.live_recap_data = self.live_recap_data[self.live_recap_data.CapLoc == cap_loc]

        # get dead mobile tracking data
        sql = 'SELECT * FROM tblMobileTracking'
        self.dead_recover_data = pd.read_sql_query(sql,con = conn, coerce_float = True)
        self.dead_recover_data = self.dead_recover_data[(self.dead_recover_data.Alive == '1') | (self.dead_recover_data.Alive == '0')]
        self.dead_recover_data['Alive'] = self.dead_recover_data.Alive.astype(np.int32)
        self.dead_recover_data.loc[self.dead_recover_data.Alive == 1,'Dead'] = 0
        self.dead_recover_data.loc[self.dead_recover_data.Alive == 0,'Dead'] = 1
        self.dead_recover_data['Dead'] = self.dead_recover_data.Dead.astype(np.int32)
        self.dead_recover_data['DateTime'] = pd.to_datetime(self.dead_recover_data.DateTime)
        self.dead_recover_data['Epoch'] = (self.dead_recover_data['DateTime'] - datetime.datetime(1970,1,1)).dt.total_seconds()

        print ("Starting receiver to recapture occasion classification, with %s records, this takes time"%(len(self.live_recap_data)))
        # create function that maps live recapture occasion to row given current receiver and apply
        def live_recap_class(row,receiver_to_recap):
            currRec = row['recID']
            recapOcc = receiver_to_recap[currRec]
            return recapOcc
        self.live_recap_data['RecapOccasion'] = self.live_recap_data.apply(live_recap_class,axis = 1, args = (receiver_to_recap,))

        # create function that maps dead recover occasion to row given current mobile tracking reach and apply
        def dead_recover_class(row,mobile_to_recap):
            currRec = row['mReach']
            recapOcc = mobile_to_recap[currRec]
            return recapOcc
        self.dead_recover_data['RecapOccasion'] = self.dead_recover_data.apply(dead_recover_class,axis = 1, args = (mobile_to_recap,))
        # get initial release states if requirec


        c.close()
        print ("Finished sql")

        if initial_recap_release == False:
            '''When initial recap release is true, the first recapture occasion
            in the rec_list query becomes our release location - this is helpful
            when modeling downstream migratory behavior of anadromous fish like
            American Shad.  We don't care about what happens on their upstream
            migration, we only care about when they turn around and come back down.

            When this argument is false, we are modeling fish from their initial
            release time and we add in R00 as our release location, otherwise R00
            should appear on the input receiver to recapture event dictionary.
            '''
            sql = "SELECT FreqCode, TagType, RelDate, RelLoc, CapLoc FROM tblMasterTag WHERE TagType = 'Study'"
            relDat = pd.read_sql_query(sql,con = conn, parse_dates = 'RelDate')
            if self.rel_loc is not None:
                relDat = relDat[relDat.RelLoc == rel_loc]
            if self.cap_loc is not None:
                relDat = relDat[relDat.CapLoc == cap_loc]
            relDat['RelDate'] = pd.to_datetime(relDat.RelDate)
            relDat['Epoch'] = (relDat['RelDate'] - datetime.datetime(1970,1,1)).dt.total_seconds()
            relDat.rename(columns = {'RelDate':'timeStamp'}, inplace = True)
            relDat['RecapOccasion'] = np.repeat('R00',len(relDat))
            #relDat.dropna(inplace =True)
            relDat['overlapping'] = np.zeros(len(relDat))
            self.data = self.data.append(relDat)
        print (self.data.head())
        c.close()
        # Identify first recapture times
        startTimes = self.data[self.data.RecapOccasion == "R00"].groupby(['FreqCode'])['Epoch'].min().to_frame()
        startTimes.reset_index(drop = False, inplace = True)
        startTimes.set_index('FreqCode',inplace = True)
        #self.startTimes.Epoch = self.startTimes.Epoch.astype(np.int32)
        startTimes.rename(columns = {'Epoch':'FirstRecapture'},inplace = True)
        print(startTimes)
        for fish in self.data.FreqCode.unique():
            if fish not in startTimes.index.values:                  # fish never made it to the initial state, remove - we only care about movements from the initial sstate - this is a competing risks model
                self.data = self.data[self.data.FreqCode != fish]
            else:
                t = startTimes.at[fish,'FirstRecapture']
                print (t)
                self.data = self.data.drop(self.data[(self.data.FreqCode == fish) & (self.data.Epoch < t)].index, inplace = True)

        # identify unique fish
        self.fish = self.live_recap_data.FreqCode.unique()                                # identify unique fish to loop through

        # store some values for safe keeping
        self.mobile_to_recap = mobile_to_recap
        self.receiver_to_recap = receiver_to_recap

        if time_limit != None:
            '''live recapture dead recovery is most useful for understanding the
            true survival of fish after they pass through infrastructure.

            often times performance criteria dictates a certain percentage of
            survival after a certain amount of time.

            If there is a time limit imposed on this performance criteria,
            enter the number of hours.  This function will first find the last
            detection in the initial state - which should be before the infrastructure -
            and will be used as passage time.  Then, this function will remove
            all detections greater than the time limit.

            Mobile tracking may only happen once per week - therefore the mobile
            tracking time limit is extended to 1 week regardless

            The time limit is expressed in decimal hours.
            '''
            # get last recapture in state 00 for all fish
            last_rec_state_00 = self.live_recap_data[self.live_recap_data.RecapOccasion == "R00"].groupby(['FreqCode'])['Epoch'].max().to_frame()
            last_rec_state_00.reset_index(drop = False, inplace = True)
            last_rec_state_00.Epoch = last_rec_state_00.Epoch.astype(np.int32)
            last_rec_state_00.rename(columns = {'Epoch':'LastRecapture_00'},inplace = True)
            print (last_rec_state_00.head())
            # join this dataframe to the live recaptures and dead recovery data
            self.live_recap_data = pd.merge(self.live_recap_data, last_rec_state_00, on = 'FreqCode', how = 'left')#, lsuffix = 'x', rsuffix = 'y')
            self.dead_recover_data = pd.merge(self.dead_recover_data, last_rec_state_00, on = 'FreqCode', how = 'left')#, lsuffix = 'x', rsuffix = 'y')
            # calculate time since last recpature in state 00
            self.live_recap_data['duration'] = self.live_recap_data.Epoch - self.live_recap_data.LastRecapture_00
            self.dead_recover_data['duration'] = self.dead_recover_data.Epoch - self.dead_recover_data.LastRecapture_00
            # convert time limit to seconds
            live_recap_time_limit = time_limit * 3600
            dead_recap_time_limit = 7 * 24 * 3600 # we only mobile track once per week - so let's increase the dead recovery time to be within a week of passage
            # extract only good data below the time limit
            self.live_recap_data = self.live_recap_data[self.live_recap_data.duration <= live_recap_time_limit]
            self.dead_recover_data = self.dead_recover_data[self.dead_recover_data.duration <= dead_recap_time_limit]

    def input_file(self,modelName, outputWS):
        # create cross tabulated data frame of live recapture data
        #Step 1: Create cross tabulated data frame with FreqCode as row index and recap occasion as column.  Identify the min epoch'''
        self.live_recap_cross_tab = pd.pivot_table(self.live_recap_data, values = 'Epoch', index = 'FreqCode', columns = 'RecapOccasion', aggfunc = 'min')
        #Step 2: Fill in those nan values with 0, we can't perform logic on nothing!'''
        self.live_recap_cross_tab.fillna(value = 0, inplace = True)
        #Step 3: Map a simply if else statement, if value > 0,1 else 0'''
        self.live_recap_cross_tab = self.live_recap_cross_tab.applymap(lambda x:1 if x > 0 else 0)
        print(self.live_recap_cross_tab.head())

        # create cross tabulated data frame of dead recovery data
        #Step 1: Create cross tabulated data frame with FreqCode as row index and recap occasion as column.  Identify the min epoch'''
        self.dead_recover_cross_tab = pd.pivot_table(self.dead_recover_data, values = 'Dead', index = 'FreqCode', columns = 'RecapOccasion', aggfunc = 'max')
        #Step 2: Fill in those nan values with 0, we can't perform logic on nothing!'''
        self.dead_recover_cross_tab.fillna(value = 0, inplace = True)
        #Step 3: Map a simply if else statement, if value > 0,1 else 0'''
        self.dead_recover_cross_tab = self.dead_recover_cross_tab.applymap(lambda x:1 if x > 0 else 0)
        print (self.dead_recover_cross_tab.head())

        # join on FreqCode
        self.inp = self.live_recap_cross_tab.join(self.dead_recover_cross_tab, on = 'FreqCode', how = 'left', lsuffix = 'x', rsuffix = 'y')
        self.inp.fillna(value = 0, inplace = True)

        # order by recapture occasion - If we start at R0, that comes after FreqCode - fuck yeah
        self.inp = self.inp.reindex(sorted(self.inp.columns), axis = 1)

        print (self.inp.head())

        # perform that LDLD logic - no alive fish after recovered dead

        # create dictionary describing inp input column and index
        inp_cols = list(self.inp.columns)
        idx = 0
        inp_col_dict = {}
        col_inp_dict = {}
        for i in inp_cols:
            inp_col_dict[i] = idx
            col_inp_dict[idx] = i
            idx = idx + 1

        col_length = len(inp_col_dict)

        # create dictionary describing mobile key index
        keys = sorted(list(self.mobile_to_recap.keys()))
        idx = 0
        mkey_idx_dict = {}
        for i in keys:
            mkey_idx_dict[i] = idx
            idx = idx + 1


        # loop through mobile reaches to find dead fish
        for i in sorted(list(self.mobile_to_recap.keys())):
            occ = self.mobile_to_recap[i]
            for index, row in self.inp.iterrows():
                row_idx = inp_col_dict[occ]
                dead = row[occ]
                if dead == 1:
                    if i != keys[-1]:
                        # this fish is dead, iterate over column indexes and write 0 there - can't be recovered alive or dead anywhere else
                        for j in np.arange(row_idx + 1,col_length,1):
                            self.inp.at[index,col_inp_dict[j]] = 0
                            row.iloc[j] = 0
                        print (row)




        print (self.inp.head())
        # Check your work
        self.inp.to_csv(os.path.join(outputWS,'%s_lrdr.csv'%(modelName)))

class time_to_event():#inputFile,outputFile,time_dependent_covariates = False, covariates = None, bucket_length = 15):
    '''Class imports standardized raw state presences and converts data structure
    into counting process style data appropriate for time to event analysis.

    Function inputs:
        input file = directory with file name of raw timestamped state presence
        output file = directory with file name for formatted data files.
        covariates = directory that contains covariates
        time_dependent_covariates = True/False (default = False) field indicating
        whether or not time dependnent covariates are incorporated into the
        counting process style
        bucket_length = covariate time interval, default 15 minutes
    '''
    def __init__(self, 
                 receiver_to_state,
                 project, 
                 input_type = 'query', 
                 initial_state_release = False, 
                 last_presence_time0 = False, 
                 cap_loc = None,
                 rel_loc = None, 
                 species = None):
        # Import Data From Project HDF
        self.rel_loc = rel_loc
        self.cap_loc = cap_loc

        # get recaptures, but first build a query
        query_parts = []
        for key in receiver_to_state:
            query_parts.append(f"rec_id == '{key}'")   
        qry = " | ".join(query_parts)

        self.recap_data = pd.read_hdf(project.db,
                                      'recaptures',
                                      where = qry)
        
        self.recap_data.drop(columns = ['power',
                                        'noise_ratio',
                                        'det_hist',
                                        'hit_ratio',
                                        'cons_det',
                                        'cons_length',
                                        'likelihood_T',
                                        'likelihood_F'],
                             axis = 'columns',
                             inplace = True)
        
        self.recap_data.set_index('freq_code', inplace = True)
        self.recap_data = pd.merge(self.recap_data,
                                   project.tags, 
                                   how = 'left',
                                   left_index = True,
                                   right_index = True)
        
        self.recap_data.reset_index(drop = False, inplace = True)
        self.recap_data.drop(columns = ['pulse_rate',
                                        'tag_type',
                                        'rel_date',
                                        'length'],
                             axis = 'columns',
                             inplace = True)
        
        # filter out tag data we don't want mucking up our staistical model
        if species != None:
            self.recap_data = self.recap_data[self.recap_data.Species == species]
        if rel_loc != None:
            self.recap_data = self.recap_data[self.recap_data.RelLoc == rel_loc]
        if cap_loc != None:
            self.recap_data = self.recap_data[self.recap_data.CapLoc == cap_loc]

        self.recap_data['state'] = self.recap_data.rec_id.map(receiver_to_state)
        
        self.recap_data = self.recap_data.astype({'freq_code':'object', 
                                                  'rec_id':'object', 
                                                  'epoch':'float32', 
                                                  'time_stamp':'datetime64', 
                                                  'lag':'float32', 
                                                  'cap_loc':'object', 
                                                  'rel_loc':'object', 
                                                  'state':'int32'})
        
        print ("Unique states in Returned data:%s"%(self.recap_data.state.unique()))

        if initial_state_release == True:
            '''we are modeling movement from the initial release location rather
            than the initial location in our competing risks model.  This allows
            us to quantify fall back.  If a fish never makes it to the intial
            spoke, then its fall back.

            If we care about modeling from the release point, we need to query
            release times of each fish, morph data into a recaptures file and
            merge it to self.data'''

            # get data   
            release_dat = project.tags
            
            # do some data management
            release_dat['rel_date'] = pd.to_datetime(release_dat.rel_date)
            release_dat['epoch'] = np.round((release_dat.rel_date - pd.Timestamp("1970-01-01")) / pd.Timedelta('1s'),6)
            release_dat.rename(columns = {'rel_date':'time_stamp'}, inplace = True)
            release_dat['rec_id'] = np.repeat('rel', len(release_dat))
            release_dat['state'] = np.zeros(len(release_dat)) 
            release_dat['presence_number'] = np.zeros(len(release_dat))
            

            # filter out tag data we don't want mucking up our staistical model
            if species != None:
                release_dat = release_dat[release_dat.Species == species]
            if rel_loc != None:
                release_dat = release_dat[release_dat.RelLoc == rel_loc]
            if cap_loc != None:
                release_dat = release_dat[release_dat.CapLoc == cap_loc]            
            
            self.recap_data = self.recap_data.append(release_dat)

        if last_presence_time0 == True:
            ''' sometimes when we are modeling downstream movement, it would be
            beneficial to start enumerating movement from the last presence at
            the initial state.

            For example, we have some long-term studies, where we are monitoring
            fish for many months.  If the fish were transported upstream for a
            downstream migration study, it may take months for that fish to move
            back downstream.  For example: Short nose sturgeon in the CT River will
            spend months to a year above Holyoke Dam potentially spawning.  They are
            not actively migrating downstream.  In this instance, it would be beneficial
            to model movement from the last presence at the initial state, after
            all, this is when it 'decides' to start its downstream migration.

            This works by finding the last presence at the initial state by fish
            and modeling movement from that point.'''

            # identify the last presence at the initial state
            self.recap_data['presence_number'] = pd.to_numeric(self.recap_data['presence_number'], 
                                                              errors='coerce')
            self.recap_data.dropna(axis = 0,
                                  how = 'any',
                                  subset = ['presence_number'], 
                                  inplace = True)
            
            last_presence = self.recap_data.groupby(['freq_code','state'])\
                ['presence_number'].max().\
                    to_frame().\
                        reset_index()
                
            last_presence.rename(columns = {'presence_number':'max_presence'},inplace = True)
            last_presence = last_presence[last_presence.state == 1]
            fish_at_start = last_presence.freq_code.unique()
            all_fish = self.recap_data.freq_code.unique()

            # now that we have the last presence, iterate over fish
            for i in last_presence.iterrows():
                fish = i[1]['freq_code']
                max_pres = i[1]['max_presence']

                # get the first recap_data at the last presence in the initial state for this fish
                recap_0 = self.recap_data[(self.recap_data.freq_code == fish) & \
                                        (self.recap_data.state == 1) & \
                                            (self.recap_data.presence_number == max_pres)]
                min_epoch = recap_0.epoch.min()

                # drop rows using boolean indexing
                self.recap_data.drop(self.recap_data[(self.recap_data.freq_code == fish) & \
                                                   (self.recap_data.epoch < min_epoch)].index, 
                                    inplace = True)
            for i in all_fish:
                if i not in fish_at_start:
                    self.recap_data.drop(self.recap_data[self.recap_data.freq_code == i].index, 
                                        inplace = True)

        # Identify first recapture times
        self.start_times = self.recap_data[self.recap_data.state == 1].\
            groupby(['freq_code'])['epoch'].min().\
                to_frame()
                    
        self.start_times.rename(columns = {'epoch':'first_recapture'},
                                inplace = True)

    	# Clean Up stuff that doesn't make sense
        for fish in self.recap_data.freq_code.unique():
            # we only care about movements from the initial sstate - this is a competing risks model
            if fish not in self.start_times.index:
                self.recap_data = self.recap_data[self.recap_data.freq_code != fish]

        # identify unique fish to loop through
        self.fish = self.recap_data.freq_code.unique()

    def data_prep(self, 
                  project,
                  time_dependent_covariates = False, 
                  unknown_state = None, 
                  bucket_length_min = 15,
                  adjacency_filter = None):
        
        if unknown_state != None:
            '''It may be beneficial to allow fish to enter into an unknown state
            rather than become censored at their last recapture in the initial state.
            This way the Nelson-Aalen will match empirical expectations.  If we have
            a lot of censored fish we lose information from the denominator and
            numerator.  If we put fish into an unknown state rather than censoring
            them we still have informative data.  For this to work, we only need to
            know the last recapture of any fish in the initial state.  We will
            assess absorbption into the unknown state with a Boolean statement later on.'''

            last_epoch = self.recap_data[self.recap_data.state == 1].epoch.max()

        if time_dependent_covariates == False:
            '''This option will produce data appropriate for construction of
            Nelson-Aalen cumulative incidence functions and to produce the state
            tables.

            This option is not appropriate if we wish to perform Cox Proportional
            Hazards Regression modeling as we will not be able to join to time-
            dependent covariates in R.
            '''
            columns = ['freq_code','state','presence','epoch','time_delta','time_0','first_obs']    # create columns
            self.master_state_table = pd.DataFrame()
            for i in self.fish:
                # get data for this fish
                fish_dat = self.recap_data[self.recap_data.freq_code == i]
                
                # get first recapture in state
                fish_dat.sort_values(by = 'epoch', 
                                     ascending = True, 
                                     inplace = True) # sort by exposure time
                
                # get previous state and fill nans with current state
                fish_dat['prev_state'] = fish_dat['state'].shift(1)  
                fish_dat.at[0,'prev_state']= fish_dat.state.values[0]
                
                # create some counters
                presence = 1
                first_obs = 1
                
                # create empty data frame
                state_table = pd.DataFrame(columns = columns)
                
                # get release time and the first epoch after release time
                time_0 = self.start_times.at[i,'first_recapture']
                fish_dat = fish_dat[fish_dat.epoch >= time_0]
                time_1 = fish_dat.epoch.iloc[0]
                
                # calculate seconds from release
                time_delta = time_1 - time_0  

                # creat a row and add it to the state table                   
                row_arr = [i,
                           fish_dat.state.values[0],
                           presence,
                           time_1,
                           time_delta,
                           time_0,
                           first_obs] # create initial row for state table
                
                row = pd.DataFrame(np.array([row_arr]),columns = columns)       
                state_table = state_table.append(row)                            
                first_obs = 0 # no other recapture can be the first
                
                # give rows an arbitrary index and find the index of our last row
                fish_dat['idx'] = np.arange(0,len(fish_dat),1)                  
                max_idx = fish_dat.idx.iloc[-1]
                
                # loop through rows, if it is a new state enter transition data                 
                for j in fish_dat.iterrows():                                   
                    row_idx = j[1]['idx']                                       
                    state = j[1]['state']                               
                    prev_state = j[1]['prev_state']  

                    if state != prev_state  or row_idx == max_idx:                
                        time_1 = j[1]['epoch']                                  
                        if unknown_state != None \
                            and row_idx == max_idx \
                                and state == 1 \
                                    and time_1 < last_epoch:
                            state = unknown_state
                            
                        time_delta = time_1 - time_0                             
                        presence = presence + 1                             
                        row_arr = [i,state,presence,time_1,time_delta,time_0,first_obs]
                        
                        row = pd.DataFrame(row_arr,columns = columns)
                        row['state'] = pd.to_numeric(row.state)
                        row = row.astype({'freq_code':'object',
                                          'state':'int32',
                                          'presence':'int32',
                                          'epoch':'float32',
                                          'time_delta':'float32',
                                          'time_0': 'float32',
                                          'first_obs':'int32'})
                        
                        state_table = state_table.append(row)                    
                        time_0 = j[1]['epoch']

                print ("State Table Completed for Fish %s"%(i))
                
                # identify transitions and write to the state table
                from_rec = state_table['state'].shift(1)
                        
                to_rec = state_table['state'].astype(np.int32)
                trans = tuple(zip(from_rec,to_rec))
                state_table['transition'] = trans
                state_table['start_state'] = from_rec
                state_table['end_state'] = to_rec
                
                # get time all sorted out
                state_table['t0'] = np.zeros(len(state_table))
                state_table['t1'] = state_table['epoch'] - state_table['time_0']
                
                # write state table to master state table
                self.master_state_table = self.master_state_table.append(state_table)

            del i,j
        else:
            columns = ['FreqCode','startState','endState','presence','time_stamp','firstObs','t0','t1']    # create columns
            
            self.master_state_table = pd.DataFrame()
            self.bucket_length = bucket_length_min
            
            for i in self.fish:
                # get fish and sort by epoch
                fish_dat = self.recap_data[self.recap_data.freq_code == i]                   # get data for this fish
                fish_dat.sort_values(by = 'epoch', 
                                     ascending = True, 
                                     inplace = True)   # sort by exposure time
                
                # identify previous state and fill in nans
                fish_dat['prev_state'] = fish_dat['state'].shift(1)                      # get previous state
                fish_dat['prev_state'].fillna(fish_dat.state.values[0], inplace = True)  # fill NaN states with current state - for first record in data frame
                
                # initialize some counters
                presence = 0
                first_obs = 1
                
                # create an empty state table 
                state_table = pd.DataFrame(columns = columns)   

                # get initial start and end times and filter dataset                 
                time_0 = self.start_times.at[i,'first_recapture'] 
                fish_dat = fish_dat[fish_dat.epoch >= time_0]
                time_1 = fish_dat.epoch.iloc[0]
                
                # calculate seconds since releaes
                time_delta = time_1 - time_0    

                # creat a row and add it to the state table                   
                row_arr = [i,
                           0,
                           fish_dat.state.values[0],
                           presence,
                           fish_dat.time_stamp.values[-1],
                           first_obs,
                           time_0,
                           time_1] # create initial row for state table
                
                row = pd.DataFrame(np.array([row_arr]),columns = columns)       
                state_table = state_table.append(row) 
                
                # create arbitrary index and get the maximum
                fish_dat['idx'] = np.arange(0,len(fish_dat),1)                  
                max_idx = fish_dat.idx.iloc[-1]                                 
                
                # for each row, if it's a new presence add data to state table
                for j in fish_dat.iterrows():                                   # for every row in fish data
                    row_idx = j[1]['idx']                                       # what's the row number?
                    state_1 = int(j[1]['prev_state'])                            # what's the state
                    state_2 = int(j[1]['state'])                                # what was the previous state
                    ts = j[1]['time_stamp']
                    
                    # if it's a new state or the end add a row
                    if state_1 != state_2 or row_idx == max_idx:                   # if the present state does not equal the previous state or if we reach the end of the dataframe...
                        time_1 = j[1]['epoch']                                  # what time is it?
                        time_delta = time_1 - time_0                              # calculate difference in seconds between current time and release                                             # if it's a new state
                        presence = presence + 1                            # oh snap new observation for new state
                        row_arr = [i,
                                   state_1,
                                   state_2,
                                   presence,
                                   ts,
                                   first_obs,
                                   time_0,
                                   time_1]  # start a new row
                        row = pd.DataFrame(np.array([row_arr]),
                                           columns = columns)
                        state_table = state_table.append(row)                    # add the row to the state table data frame
                        time_0 = j[1]['epoch']
                        first_obs = 0
                        
                print ("State Table Completed for Fish %s"%(i))

                state_table.sort_values(by = 'time_0', 
                                        ascending = True, 
                                        inplace = True) # sort by exposure time
                
                # put time into increments to match time series variables
                time_bucket = self.bucket_length*60*1000000000                  # time bucket in nanoseconds
                state_table['flow_period'] = (state_table['time_0'].\
                                              astype(np.int64)//time_bucket+1) * time_bucket # round to nearest 15 minute period
                state_table['flow_period'] = pd.to_datetime(state_table['flow_period']) # turn it into a datetime object so we can use pandas to expand and fill
                
                # create arbitrary index
                row_num = np.arange(0,len(state_table),1)
                state_table['row_num'] = row_num
                
                # create an expanded state table
                exp_state_table = pd.DataFrame()
                
                # build expanded state table
                for row in state_table.iterrows():
                    row_idx = row[1]['row_num']                                 
                    t0 = row[1]['flow_period']                                
                    t1 = row[1]['t1']                                         
                    
                    # try expanding, if interval not large enough return nothing 
                    try:
                        expand = pd.date_range(t0,
                                               t1,
                                               freq = '%smin'%(self.bucket_length)) 
                    except ValueError:
                        expand = []
                    except AttributeError:
                        expand = []
                        
                    # if we can expand create intervals 
                    if len(expand) > 0:
                        # create a series using expanded time stamps
                        series = pd.Series(expand, 
                                           index = expand, 
                                           name = 'flow_period') 
                        
                        # convert series to invterval dataframe and perform data management 
                        intervals = series.to_frame()                          
                        intervals.reset_index(inplace = True, drop = True)     
                        intervals['t0'] = row[1]['t0']                         
                        intervals['t1'] = row[1]['t1']
                        intervals['startState'] = row[1]['startState']
                        intervals['endState'] = row[1]['endState']
                        intervals['timeStamp'] = row[1]['timeStamp']
                        intervals['FreqCode'] = row[1]['FreqCode']
                        intervals['presence'] = row[1]['presence']
                        newRowArr = np.array([row[1]['FreqCode'],
                                              row[1]['startState'],
                                              row[1]['endState'],
                                              row[1]['timeStamp'],
                                              row[1]['flowPeriod'],
                                              row[1]['t0'],
                                              row[1]['t1'],
                                              row[1]['presence']])
                        newRow = pd.DataFrame(np.array([newRowArr]),columns = ['FreqCode',
                                                                               'startState',
                                                                               'endState',
                                                                               'timeStamp',
                                                                               'flowPeriod',
                                                                               't0',
                                                                               't1',
                                                                               'presence']) # add first, non expanded row to new state table
                        newRow = newRow.append(intervals)                      # add filled and expanded data
                        newRow['nextFlowPeriod'] = newRow['flowPeriod'].shift(-1) # identify the next flow period
                        newRow['idx'] = np.arange(0,len(newRow),1)             # add a count index field, but don't index it yet
                        newRow.reset_index(inplace = True, drop = True)        # remove the index
                        idxL = newRow.idx.values                               # generate list of indexes
                        newRow.loc[idxL[1]:,'t0'] = newRow.loc[idxL[1]:,'flowPeriod'].astype(str) # after the initial t0, re-write the current t0 as the current row's flow period
                        newRow.ix[:idxL[-2],'t1'] = newRow.loc[:idxL[-2],'nextFlowPeriod'].astype(str)  # other than the last t1, re-write the current t1 as the current row's next flow period - see what we did there?
                        newRow.ix[:idxL[-2]:,'endState'] = row[1]['startState']# other than the last row in the series, re-write the end state as the start state - there will be a lot of to-from same site here. it's ok, these are censored observations.
                        newRow['t0'] = pd.to_datetime(newRow['t0'])            # convert time text to datetime - so we can do stuff with it
                        newRow['t1'] = pd.to_datetime(newRow['t1'])
                        exp_state_table = exp_state_table.append(newRow)         # now add all that stuff to the state table dataframe
                        del newRow, intervals, newRowArr, expand
                    else:
                        newRowArr = np.array([row[1]['FreqCode'],
                                              row[1]['startState'],
                                              row[1]['endState'],
                                              row[1]['timeStamp'],
                                              row[1]['flowPeriod'],
                                              row[1]['t0'],
                                              row[1]['t1'],
                                              row[1]['presence']])
                        newRow = pd.DataFrame(np.array([newRowArr]),columns = ['FreqCode',
                                                                               'startState',
                                                                               'endState',
                                                                               'timeStamp',
                                                                               'flowPeriod',
                                                                               't0',
                                                                               't1',
                                                                               'presence']) # add first, non expanded row to new state table
                        exp_state_table = exp_state_table.append(newRow)
                        del newRow, newRowArr
                # exp_state_table.sort_values(by = 't0', ascending = True, inplace = True)     # sort by exposure time
                # exp_state_table['time0'] = pd.to_datetime(exp_state_table['t0']) # create new time columns
                # exp_state_table['time1'] = pd.to_datetime(exp_state_table['t1'])
                # exp_state_table['t0'] = (pd.to_datetime(exp_state_table['t0']) - initialTime)/np.timedelta64(1,'s')
                # exp_state_table['t1'] = (pd.to_datetime(exp_state_table['t1']) - initialTime)/np.timedelta64(1,'s')
                # # calculate minimum t0 by presence
                # min_t0 = exp_stateTable.groupby(['presence'])['t0'].min()#.to_frame().rename({'t0':'min_t0'},inplace = True)
                # min_t0 = pd.Series(min_t0, name = 'min_t0')
                # min_t0 = pd.DataFrame(min_t0).reset_index()
                # # join to exp_stateTable as presence_time_0
                # exp_stateTable = pd.merge(left = exp_stateTable, right = min_t0, how = u'left',left_on = 'presence', right_on = 'presence')
                # # subtract presence_time_0 from t0 and t1
                # exp_stateTable['t0'] = exp_stateTable['t0'] -  exp_stateTable['min_t0']
                # exp_stateTable['t1'] = exp_stateTable['t1'] -  exp_stateTable['min_t0']
                # # drop presence_time_0 from exp_stateTable

                # exp_stateTable['hour'] = pd.DatetimeIndex(exp_stateTable['time0']).hour # get the hour of the day from the current time stamp
                # exp_stateTable['qDay'] = exp_stateTable.hour//6                # integer division by 6 to put the day into a quarter
                # exp_stateTable['test'] = exp_stateTable.t1 - exp_stateTable.t0 # this is no longer needed, but if t1 is smaller than t0 things are screwed up
                # stateTable = exp_stateTable
                # del exp_stateTable
                # stateTable['transition'] = tuple(zip(stateTable.startState.values.astype(int),stateTable.endState.values.astype(int))) # create transition variable, this is helpful in R
                # self.master_stateTable = self.master_stateTable.append(stateTable)
                # export
            self.master_stateTable.drop(labels = ['nextFlowPeriod'],axis = 1, inplace = True)


        if adjacency_filter is not None:
            '''When the truth value of a detection is assessed, a detection
            may be valid for a fish that is not present.

            In some instances, especially when using Yagi antennas, back-lobes
            may develop where a fish in the tailrace of a powerhouse is
            detected in the forebay antenna.  In these instances, a downstream
            migrating fish would not have migrated up through the powerhouse.

            From a false positive perspective, these records are valid detections.
            However, from a movement perspective, these series of detections
            could not occur and should be removed.

            This function repeatedly removes rows with 'illegal' movements
            until there are none left.  Rows with 'illegal' transitions are
            identified with a list that is passed to the function.

            input = list of illegal transitions stored as (from, to) tuples
            '''
            fish = self.master_stateTable.FreqCode.unique()

            for i in fish:
                fishDat =  self.master_stateTable[self.master_stateTable.FreqCode == i]
                self.master_stateTable = self.master_stateTable[self.master_stateTable.FreqCode != i]

                # create a condition, we're running this filter because we know illogical movements are present
                bad_moves_present = True

                # while there are illogical movements, keep filtering
                while bad_moves_present == True:
                    # let's keep count of the number of rows we are filtering
                    filtered_rows = 0.0

                    # for every known bad movement
                    for j in adjacency_filter:
                        print ("Starting %s filter"%(i))
                        # find those rows where this movement exists
                        fishDat['transition_filter'] = np.where(fishDat.transition == j,1,0)
                        fishDat.set_index(['time0'], inplace = True)

                        if fishDat.transition_filter.sum() > 0:
                            # add up those rows
                            filtered_rows = filtered_rows + fishDat.transition_filter.sum()
                            print ('%s rows found with %s movements'%(fishDat.transition_filter.sum(),j))

                            # do some data management, we need to take the start state and t0 of the affected rows and place them on the subsequent row
                            idx = fishDat.index[fishDat['transition_filter']==1]

                            for k in idx:
                                idx_int = fishDat.index.get_loc(k)
                                t0_col = fishDat.columns.get_loc('t0')
                                start_col = fishDat.columns.get_loc('startState')

                                # get start time and start state
                                start = fishDat.iloc[idx_int]['startState']
                                t0 = fishDat.iloc[idx_int]['t0']

                                # write it to next row
                                try:
                                    idx1 = idx_int + 1
                                except:
                                    start = fishDat.iloc[idx_int].index[0]
                                    idx1 = start + 1
                                try:
                                    fishDat.iloc[idx1, start_col] = start
                                    fishDat.iloc[idx1, t0_col] = t0
                                except IndexError:
                                    # when this occurs, there is no extra row - this last row will be deleted
                                    continue

                            # remove those rows
                            fishDat = fishDat[fishDat.transition_filter != 1]

                            # create a new transition field
                            fishDat['transition'] = tuple(zip(fishDat.startState.values.astype(int),
                                                              fishDat.endState.values.astype(int)))

                            fishDat.reset_index(inplace = True)
                        else:
                            print ("No illegal movements identified")
                            fishDat.reset_index(inplace = True)

                    if filtered_rows == 0.0:
                        print ("All illegal movements for fish %s removed"%(i))
                        # stop that loop
                        bad_moves_present = False

                    else:
                        # i feel bad for you son
                        print ("%s illegal movements present in iteration, go again"%(filtered_rows))

                fishDat.drop(labels = ['transition_filter'], axis = 1, inplace = True)
                self.master_stateTable = self.master_stateTable.append(fishDat)

        #self.master_stateTable = self.master_stateTable[self.master_stateTable.firstObs == 0]
        self.master_stateTable.to_csv(outputFile)

    # generate summary statistics
    def summary(self):
        print ("--------------------------------------------------------------------------------------------------------------")
        print ("Time To Event Data Manage Complete")
        print ("--------------------------------------------------------------------------------------------------------------")
        print ("")
        print ("")
        print ("---------------------------------------MOVEMENT SUMMARY STATISTICS--------------------------------------------")
        print ("")
        print ("In Total, there were %s unique fish within this competing risks model"%(len(self.master_stateTable.FreqCode.unique())))
        print ("The number of unique fish per state:")
        countPerState = self.master_stateTable.groupby(['state'])['FreqCode'].nunique().to_frame()
        print (countPerState)
        print ("")
        msm_stateTable = pd.crosstab(self.master_stateTable.startState, self.master_stateTable.endState)
        print ("These fish made the following movements as enumerated in the state transition table:")
        print (msm_stateTable)
        print ("The table should read movement from a row to a column")
        print ("")
        self.master_stateTable['transition'] = self.master_stateTable.transition.astype(str)
        self.countTable = self.master_stateTable.groupby(['startState','endState'])['FreqCode'].nunique().to_frame()
        self.countTable.reset_index(inplace = True)
        countPerTrans = pd.crosstab(self.countTable.startState,self.countTable.endState,values = self.countTable.FreqCode, aggfunc = 'sum')
        print ("The number of unique fish to make these movements are found in the following count table")
        print (countPerTrans)
        print ("")
        # Step 3: Describe the expected number of transitions per fish
        self.fishTransCount = self.master_stateTable.groupby(['FreqCode','transition'])['transition'].count()
        self.fishTransCount = self.fishTransCount.to_frame(name = 'transCount')
        #self.fishTransCount.rename(columns = {'':'transCount'}, inplace = True)
        #self.fishTransCount.reset_index(inplace = True)

        print ("The number of movements a fish is expected to make is best described with min, median and maximum statistics")
        print ("The mininum number of times each transition was made:")
        min_transCount = self.fishTransCount.groupby(['transition'])['transCount'].min()
        print (min_transCount)
        print ("")
        print ("The median number of times each transition was made:")
        med_transCount = self.fishTransCount.groupby(['transition'])['transCount'].median()
        print (med_transCount)
        print ("")
        print ("The maximum number of times each transition was made by each fish:")
        max_transCount = self.fishTransCount.groupby(['FreqCode','transition'])['transCount'].max()
        #max_transCount.to_csv(os.path.join(r'C:\Users\Kevin Nebiolo\Desktop','maxCountByFreqCode.csv'))
        print (max_transCount)
        print ("")
        print ("Movement summaries - Duration between states in seconds")
        self.master_stateTable['dur'] = (self.master_stateTable.t1 - self.master_stateTable.t0)
        move_summ = self.master_stateTable.groupby('transition')['dur'].describe().round(decimals = 3)
        print (move_summ)



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 15 21:50:12 2023

@author: KNebiolo
"""
import numpy as np

def calculate_priors(labeled_array):
    """
    Calculate the prior probabilities for a binary classification problem.

    Parameters:
    labeled_array: A numpy array of boolean values (True/False) representing the truth value of detections.

    Returns:
    A tuple containing the prior probabilities (prior_true, prior_false).
    """
    total_count = labeled_array.size
    true_count = np.sum(labeled_array)
    false_count = total_count - true_count

    prior_true = true_count / total_count
    prior_false = false_count / total_count

    return (prior_true, prior_false)

def calculate_likelihood(training_obs, labeled_array, assumption, classification_obs, laplace=1):
    """
    Calculate the likelihood of each value in the classification dataset given a 
    truth value (True or False), based on training data with Laplace smoothing.

    Parameters:
    training_obs: A numpy array of observed values (binned data) from training data.
    labeled_array: A numpy array of boolean values (True/False) from training data, 
    representing the truth value.
    assumption: The truth value (True or False) for which to calculate the likelihood.
    classification_obs: A numpy array of observed values (binned data) from 
    classification dataset.
    laplace: The smoothing parameter (default is 1 for add-one smoothing).

    Returns:
    A 1D numpy array containing the likelihood of each observed value in the 
    lassification dataset given the assumption.
    """
    # Filter the training observation array based on the assumption
    filtered_obs = training_obs[labeled_array == assumption]

    # Count occurrences of each value in the filtered observation array
    unique_values, counts = np.unique(filtered_obs, return_counts=True)

    # Get the total number of unique values in the training observations
    total_unique_values = len(np.unique(training_obs))

    # Calculate the likelihood for each unique value with Laplace smoothing
    likelihoods = {val: (count + laplace) / (len(filtered_obs) + laplace * total_unique_values) for val, count in zip(unique_values, counts)}

    # Default likelihood for unseen values after smoothing
    default_likelihood = laplace / (len(filtered_obs) + laplace * total_unique_values)

    # Map each observation in the classification dataset to its corresponding likelihood
    likelihood_array = np.array([likelihoods.get(obs, default_likelihood) for obs in classification_obs])

    return likelihood_array

def calculate_evidence(observation_arrays):
    """
    Calculate the evidence for a Naive Bayes classifier with multiple predictors.

    Parameters:
    *observation_arrays: variable number of arrays, each representing binned data for a predictor.

    Returns:
    A 1D numpy array containing the evidence for each record.
    """

    # Calculate the marginal probabilities for each bin in each predictor
    marginal_probs = [np.bincount(obs_array) / len(obs_array) for obs_array in observation_arrays]

    # Calculate the evidence for each record
    evidence = np.ones(len(observation_arrays[0]))
    for obs_array, probs in zip(observation_arrays, marginal_probs):
        evidence *= probs[obs_array.astype(int)]

    return evidence

def calculate_posterior(priors, evidence, likelihoods_dict, assumption):
    """
    Calculate the posterior probabilities given observations.

    Parameters:
    priors: A 1D numpy array containing the prior probabilities for True and False.
    evidence: A 1D numpy array containing the evidence for each instance.
    likelihoods_dict: A nested dictionary of the structure {assumption:{predictor:likelihood array}}.
    assumption: The truth value (True or False) for which to calculate the posterior.

    Returns:
    A 1D numpy array containing the posterior probabilities for each instance.
    """
    # Initialize the posterior array with the prior probabilities
    posterior = np.full_like(evidence, priors[assumption])

    # Multiply by the likelihoods for each predictor
    for predictor, likelihood_array in likelihoods_dict[assumption].items():
        posterior *= likelihood_array

    # Divide by the evidence
    posterior /= evidence

    return posterior

def classify_with_threshold(posterior_true, posterior_false, threshold_ratio=1.0):
    """
    Classify instances based on posterior probabilities with an adjustable threshold.

    Parameters:
    posterior_true: A numpy array containing the posterior probabilities of being True.
    posterior_false: A numpy array containing the posterior probabilities of being False.
    threshold_ratio: The ratio of posterior_true to posterior_false required to 
                     classify an instance as True.
                     Default is 1.0, which corresponds to the standard MAP approach.

    Returns:
    A numpy array of classifications (True or False).
    """
    # Calculate the ratio of posterior_true to posterior_false
    ratio = posterior_true / posterior_false

    # Classify as True where the ratio exceeds the threshold, otherwise False
    classification = ratio >= threshold_ratio

    return classification

def bin_predictors(hit_ratio, power, lag_diff, cons_length, noise_ratio):
    'bin numerical predictors for classification'
    # define bins for analysis
    
    # hit ratio bins
    hit_ratio_bins =np.linspace(0,1.0,11)

    # plot signal power histograms by detection class
    min_power = power.min()//5 * 5
    max_power = power.max()//5 * 5
    try:
        power_bins =np.arange(min_power,max_power+20,10)
    except:
        print ('fuck')

    # Lag Back Differences - how steady are detection lags?
    lag_bins =np.arange(-100,110,20)

    # Consecutive Record Length
    con_length_bins =np.arange(1,12,1)

    # Noise Ratio
    noise_bins =np.arange(0,1.1,0.1)
    
    # use np.digitize to bind
    hit_ratio_count = np.digitize(hit_ratio,  hit_ratio_bins)
    power_count = np.digitize(power,power_bins)
    lag_count = np.digitize(lag_diff, lag_bins)
    con_len_count = np.digitize(cons_length, con_length_bins)
    noise_count = np.digitize(noise_ratio, noise_bins)
    
    return hit_ratio_count, power_count, lag_count, con_len_count, noise_count